@@ -1,6 +1,6 @@
-// $Id: CommandLine.cpp,v 1.41 2004/05/31 18:36:20 geuzaine Exp $
+// $Id: CommandLine.cpp,v 1.69.2.1 2006/03/17 21:16:36 geuzaine Exp $
 //
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -19,7 +19,6 @@
 // 
 // Please report all bugs and problems to <gmsh@geuz.org>.
 
-#include <unistd.h>
 #include "Gmsh.h"
 #include "GmshUI.h"
 #include "GmshVersion.h"
@@ -32,6 +31,7 @@
 #include "Views.h"
 #include "OpenFile.h"
 #include "Parser.h"
+#include "OS.h"
 
 #if !defined(GMSH_EXTRA_VERSION)
 #error
@@ -43,13 +43,8 @@
 extern Context_T CTX;
 extern Mesh *THEM;
 
-char *TheFileNameTab[MAX_OPEN_FILES];
-char *TheBgmFileName = NULL, *TheOptString = NULL;
-
-// *INDENT-OFF*
-
 char gmsh_progname[]  = "Gmsh, a 3D mesh generator with pre- and post-processing facilities" ;
-char gmsh_copyright[] = "Copyright (C) 1997-2004 Christophe Geuzaine and Jean-Francois Remacle";
+char gmsh_copyright[] = "Copyright (C) 1997-2006 Christophe Geuzaine and Jean-Francois Remacle";
 char gmsh_version[]   = "Version        : " ;
 char gmsh_license[]   = "License        : " GMSH_SHORT_LICENSE;
 char gmsh_gui[]       = "GUI toolkit    : " ;
@@ -61,56 +56,69 @@ char gmsh_packager[]  = "Packager       : " GMSH_PACKAGER ;
 char gmsh_url[]       = "Web site       : http://www.geuz.org/gmsh/" ;
 char gmsh_email[]     = "Mailing list   : gmsh@geuz.org" ;
 
-// If you change the following, please also change
-// * the texinfo documentation (doc/texinfo/command_line.texi)
-// * the man page (doc/gmsh.1)
+int Get_GmshMajorVersion(){ return GMSH_MAJOR_VERSION; }
+int Get_GmshMinorVersion(){ return GMSH_MINOR_VERSION; }
+int Get_GmshPatchVersion(){ return GMSH_PATCH_VERSION; }
+char *Get_GmshExtraVersion(){ return GMSH_EXTRA_VERSION; }
+char *Get_GmshVersion(){ return GMSH_VERSION; }
+char *Get_GmshBuildDate(){ return GMSH_DATE; }
+char *Get_GmshBuildHost(){ return GMSH_HOST; }
+char *Get_GmshPackager(){ return GMSH_PACKAGER; }
+char *Get_GmshBuildOS(){ return GMSH_OS; }
+char *Get_GmshShortLicense(){ return GMSH_SHORT_LICENSE; }
 
 void Print_Usage(char *name){
+  // If you make changes in this routine, please also change the
+  // texinfo documentation (doc/texinfo/command_line.texi) as well as
+  // the man page (doc/gmsh.1)
   Msg(DIRECT, "Usage: %s [options] [files]", name);
   Msg(DIRECT, "Geometry options:");
-  Msg(DIRECT, "  -0                    parse input files, output unrolled geometry, and exit");
+  Msg(DIRECT, "  -0                    Parse input files, output unrolled geometry, and exit");
   Msg(DIRECT, "Mesh options:");
-  Msg(DIRECT, "  -1, -2, -3            perform batch 1D, 2D and 3D mesh generation");
-  Msg(DIRECT, "  -saveall              save all elements (discard physical group definitions)");
-  Msg(DIRECT, "  -o file               specify mesh output file name");
-  Msg(DIRECT, "  -format msh|unv|gref  set output mesh format (default: msh)");
-  Msg(DIRECT, "  -algo iso|tri|aniso   select 2D mesh algorithm (default: iso)");
-  Msg(DIRECT, "  -smooth int           set mesh smoothing (default: 0)");
-  Msg(DIRECT, "  -order int            set mesh order (default: 1)");
-  Msg(DIRECT, "  -scale float          set global scaling factor (default: 1.0)");
-  Msg(DIRECT, "  -meshscale float      set mesh scaling factor (default: 1.0)");
-  Msg(DIRECT, "  -clscale float        set characteristic length scaling factor (default: 1.0)");
-  Msg(DIRECT, "  -rand float           set random perturbation factor (default: 1.e-4)");
-  Msg(DIRECT, "  -bgm file             load backround mesh from file");
-  Msg(DIRECT, "  -constrain            constrain background mesh with characteristic lengths");
-  Msg(DIRECT, "  -histogram            print mesh quality histogram");
-  Msg(DIRECT, "  -extrude              use old extrusion mesh generator");
-  Msg(DIRECT, "  -recombine            recombine meshes from old extrusion mesh generator");
+  Msg(DIRECT, "  -1, -2, -3            Perform batch 1D, 2D and 3D mesh generation");
+  Msg(DIRECT, "  -saveall              Save all elements (discard physical group definitions)");
+  Msg(DIRECT, "  -o file               Specify mesh output file name");
+  Msg(DIRECT, "  -format string        Set output mesh format (msh, unv, gref, stl, p3d)");
+  Msg(DIRECT, "  -algo string          Select mesh algorithm (iso, tri, aniso, netgen, tetgen)");
+  Msg(DIRECT, "  -smooth int           Set number of mesh smoothing steps");
+  Msg(DIRECT, "  -optimize             Optimize quality of tetrahedral elements");
+  Msg(DIRECT, "  -partition int        Load a mesh and partition it into $1 parts (METIS is required)");
+  Msg(DIRECT, "  -order int            Set mesh order (1, 2)");
+  Msg(DIRECT, "  -scale float          Set global scaling factor");
+  Msg(DIRECT, "  -meshscale float      Set mesh scaling factor");
+  Msg(DIRECT, "  -clscale float        Set characteristic length scaling factor");
+  Msg(DIRECT, "  -rand float           Set random perturbation factor");
+  Msg(DIRECT, "  -bgm file             Load background mesh from file");
+  Msg(DIRECT, "  -constrain            Constrain background mesh with characteristic lengths");
+  Msg(DIRECT, "  -histogram            Print mesh quality histogram");
+  Msg(DIRECT, "  -extrude              Use old extrusion mesh generator");
+  Msg(DIRECT, "  -recombine            Recombine meshes from old extrusion mesh generator");
 #if defined(HAVE_FLTK)
-  Msg(DIRECT, "  -interactive          display 2D mesh construction interactively");
+  Msg(DIRECT, "  -interactive          Display 2D mesh construction interactively");
   Msg(DIRECT, "Post-processing options:");
-  Msg(DIRECT, "  -noview               hide all views on startup");
-  Msg(DIRECT, "  -link int             select link mode between views (default: 0)");
-  Msg(DIRECT, "  -smoothview           smooth views");
-  Msg(DIRECT, "  -combine              combine input views into multi time step ones");
+  Msg(DIRECT, "  -noview               Hide all views on startup");
+  Msg(DIRECT, "  -link int             Select link mode between views (0, 1, 2, 3, 4)");
+  Msg(DIRECT, "  -combine              Combine views having identical names into multi-time-step views");
   Msg(DIRECT, "Display options:");    
-  Msg(DIRECT, "  -nodb                 disable double buffering");
-  Msg(DIRECT, "  -fontsize int         specify the font size for the GUI (default: 12)");
-  Msg(DIRECT, "  -scheme string        specify FLTK GUI scheme");
-  Msg(DIRECT, "  -display string       specify display");
+  Msg(DIRECT, "  -nodb                 Disable double buffering");
+  Msg(DIRECT, "  -fontsize int         Specify the font size for the GUI");
+  Msg(DIRECT, "  -scheme string        Specify FLTK GUI scheme");
+  Msg(DIRECT, "  -display string       Specify display");
 #endif
   Msg(DIRECT, "Other options:");      
 #if defined(HAVE_FLTK)
-  Msg(DIRECT, "  -a, -g, -m, -s, -p    start in automatic, geometry, mesh, solver or post-processing");
-  Msg(DIRECT, "                        mode (default: automatic)");
+  Msg(DIRECT, "  -a, -g, -m, -s, -p    Start in automatic, geometry, mesh, solver or post-processing mode");
 #endif
-  Msg(DIRECT, "  -v int                set verbosity level (default: 3)");
-  Msg(DIRECT, "  -string \"string\"      parse string before project file");
-  Msg(DIRECT, "  -option file          parse option file before GUI creation");
-  Msg(DIRECT, "  -convert file file    perform batch conversion of view(s)/mesh into latest file formats");
-  Msg(DIRECT, "  -version              show version number");
-  Msg(DIRECT, "  -info                 show detailed version information");
-  Msg(DIRECT, "  -help                 show this message");
+  Msg(DIRECT, "  -pid                  Print process id on stdout");
+  Msg(DIRECT, "  -listen               Always listen to incoming connection requests");
+  Msg(DIRECT, "  -v int                Set verbosity level");
+  Msg(DIRECT, "  -nopopup              Don't popup dialog windows in scripts");
+  Msg(DIRECT, "  -string \"string\"      Parse option string at startup");
+  Msg(DIRECT, "  -option file          Parse option file at startup");
+  Msg(DIRECT, "  -convert file file    Perform batch conversion of views and meshes into latest file formats");
+  Msg(DIRECT, "  -version              Show version number");
+  Msg(DIRECT, "  -info                 Show detailed version information");
+  Msg(DIRECT, "  -help                 Show this message");
 }
 
 char *Get_BuildOptions(void)
@@ -125,6 +133,12 @@ char *Get_BuildOptions(void)
 #if defined(HAVE_TRIANGLE)
     strcat(opt, "TRIANGLE ");
 #endif
+#if defined(HAVE_NETGEN)
+    strcat(opt, "NETGEN ");
+#endif
+#if defined(HAVE_TETGEN)
+    strcat(opt, "TETGEN ");
+#endif
 #if defined(HAVE_LIBJPEG)
     strcat(opt, "JPEG ");
 #endif
@@ -136,45 +150,34 @@ char *Get_BuildOptions(void)
 #endif
 #if defined(HAVE_MATH_EVAL)
     strcat(opt, "MATHEVAL ");
+#endif
+#if defined(HAVE_METIS)
+    strcat(opt, "METIS ");
+#endif
+#if defined(HAVE_ANN_)
+    strcat(opt, "ANN ");
 #endif
     first = 0;
   }
   return opt;
 }
 
-// *INDENT-ON*
-
-void Get_Options(int argc, char *argv[], int *nbfiles)
+void Get_Options(int argc, char *argv[])
 {
-  int i = 1;
-
-  // This symbol context is local to option parsing (the symbols will
-  // not interfere with subsequent OpenFiles)
-
-  InitSymbols();
-
   // Parse session and option files
-
-  ParseFile(CTX.sessionrc_filename, 1, 1);
-  ParseFile(CTX.optionsrc_filename, 1, 1);
+  ParseFile(CTX.session_filename_fullpath, 1, 1);
+  ParseFile(CTX.options_filename_fullpath, 1, 1);
 
   // Get command line options
-
-  TheFileNameTab[0] = CTX.default_filename;
-  *nbfiles = 0;
-
+  int i = 1;
   while(i < argc) {
 
     if(argv[i][0] == '-') {
 
-      if(!strcmp(argv[i] + 1, "string")) {
+      if(!strcmp(argv[i] + 1, "pid")) {
+	fprintf(stdout, "%d\n", GetProcessId());
+	fflush(stdout);
         i++;
-        if(argv[i] != NULL)
-          TheOptString = argv[i++];
-        else {
-          fprintf(stderr, ERROR_STR "Missing string\n");
-          exit(1);
-        }
       }
       else if(!strcmp(argv[i] + 1, "a")) {
         CTX.initial_context = 0;
@@ -232,6 +235,23 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
         CTX.mesh.histogram = 1;
         i++;
       }
+      else if(!strcmp(argv[i] + 1, "optimize")) {
+        CTX.mesh.optimize = 1;
+        i++;
+      }
+      else if(!strcmp(argv[i] + 1, "nopopup")) {
+        CTX.nopopup = 1;
+        i++;
+      }
+      else if(!strcmp(argv[i] + 1, "string")) {
+        i++;
+        if(argv[i] != NULL)
+	  ParseString(argv[i++]);
+        else {
+          fprintf(stderr, ERROR_STR "Missing string\n");
+          exit(1);
+        }
+      }
       else if(!strcmp(argv[i] + 1, "option")) {
         i++;
         if(argv[i] != NULL)
@@ -253,7 +273,7 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
       else if(!strcmp(argv[i] + 1, "bgm")) {
         i++;
         if(argv[i] != NULL)
-          TheBgmFileName = argv[i++];
+          CTX.bgm_filename = argv[i++];
         else {
           fprintf(stderr, ERROR_STR "Missing file name\n");
           exit(1);
@@ -271,7 +291,7 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
           OpenProblem(argv[i]);
 	  // convert post-processing views to latest (binary) format
           for(int j = 0; j < List_Nbr(CTX.post.list); j++)
-            WriteView((Post_View *) List_Pointer(CTX.post.list, j),
+            WriteView(*(Post_View **) List_Pointer(CTX.post.list, j),
                       argv[i + 1], 1, j ? 1 : 0);
 	  // convert any mesh to the latest format
 	  if(THEM){
@@ -285,10 +305,6 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
           fprintf(stderr, "Usage: %s -convert file file\n", argv[0]);
         exit(1);
       }
-      else if(!strcmp(argv[i] + 1, "old")) {
-        CTX.geom.old_circle = 1;
-        i++;
-      }
       else if(!strcmp(argv[i] + 1, "initial")) {
         i++;
         if(argv[i] != NULL)
@@ -349,6 +365,16 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
           exit(1);
         }
       }
+      else if(!strcmp(argv[i] + 1, "partition")) {
+        i++;
+        if(argv[i] != NULL) {
+          CTX.mesh.nbPartitions = atoi(argv[i++]);
+        }
+        else {
+          fprintf(stderr, ERROR_STR "Missing number\n");
+          exit(1);
+        }
+      }
       else if(!strcmp(argv[i] + 1, "smooth")) {
         i++;
         if(argv[i] != NULL)
@@ -382,6 +408,13 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
                   !strcmp(argv[i], "GREF") || !strcmp(argv[i], "Gref")) {
             CTX.mesh.format = FORMAT_GREF;
           }
+          else if(!strcmp(argv[i], "stl") || !strcmp(argv[i], "STL")) {
+            CTX.mesh.format = FORMAT_STL;
+          }
+          else if(!strcmp(argv[i], "p3d") ||
+                  !strcmp(argv[i], "P3D") || !strcmp(argv[i], "Plot3D")) {
+            CTX.mesh.format = FORMAT_P3D;
+          }
           else {
             fprintf(stderr, ERROR_STR "Unknown mesh format\n");
             exit(1);
@@ -397,11 +430,15 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
         i++;
         if(argv[i] != NULL) {
           if(!strncmp(argv[i], "iso", 3))
-            CTX.mesh.algo = DELAUNAY_ISO;
+            CTX.mesh.algo2d = DELAUNAY_ISO;
           else if(!strncmp(argv[i], "tri", 3))
-            CTX.mesh.algo = DELAUNAY_SHEWCHUK;
+            CTX.mesh.algo2d = DELAUNAY_TRIANGLE;
           else if(!strncmp(argv[i], "aniso", 5))
-            CTX.mesh.algo = DELAUNAY_ANISO;
+            CTX.mesh.algo2d = DELAUNAY_ANISO;
+          else if(!strncmp(argv[i], "netgen", 6))
+            CTX.mesh.algo3d = FRONTAL_NETGEN;
+          else if(!strncmp(argv[i], "tetgen", 6))
+            CTX.mesh.algo3d = DELAUNAY_TETGEN;
           else {
             fprintf(stderr, ERROR_STR "Unknown mesh algorithm\n");
             exit(1);
@@ -413,8 +450,11 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
           exit(1);
         }
       }
-      else if(!strcmp(argv[i] + 1, "version") ||
-              !strcmp(argv[i] + 1, "-version")) {
+      else if(!strcmp(argv[i] + 1, "listen")) {
+        CTX.solver.listen = 1;
+        i++;
+      }
+      else if(!strcmp(argv[i] + 1, "version") || !strcmp(argv[i] + 1, "-version")) {
         fprintf(stderr, "%s\n", GMSH_VERSION);
         exit(1);
       }
@@ -478,10 +518,6 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
           exit(1);
         }
       }
-      else if(!strcmp(argv[i] + 1, "fill")) {
-        opt_view_intervals_type(0, GMSH_SET, DRAW_POST_CONTINUOUS);
-        i++;
-      }
       else if(!strcmp(argv[i] + 1, "smoothview")) {
         CTX.post.smooth = 1;
         i++;
@@ -490,27 +526,6 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
         CTX.post.combine_time = 1;
         i++;
       }
-      else if(!strcmp(argv[i] + 1, "nbiso")) {
-        i++;
-        if(argv[i] != NULL)
-          opt_view_nb_iso(0, GMSH_SET, atoi(argv[i++]));
-        else {
-          fprintf(stderr, ERROR_STR "Missing number\n");
-          exit(1);
-        }
-      }
-      else if(!strcmp(argv[i] + 1, "threads")) {
-        CTX.threads = 1;
-        i++;
-      }
-      else if(!strcmp(argv[i] + 1, "nothreads")) {
-        CTX.threads = 0;
-        i++;
-      }
-      else if(!strcmp(argv[i] + 1, "db")) {
-        CTX.db = 1;
-        i++;
-      }
       else if(!strcmp(argv[i] + 1, "nodb")) {
         CTX.db = 0;
         i++;
@@ -549,31 +564,30 @@ void Get_Options(int argc, char *argv[], int *nbfiles)
         }
       }
 #endif
-      else {
 #if defined(__APPLE__)
-        // The Mac Finder launches programs with special command line
-        // arguments: just ignore them (and don't exit!)
-        fprintf(stderr, "Unknown option '%s'\n", argv[i]);
+      else if(!strncmp(argv[i] + 1, "psn", 3)) {
+        // The Mac Finder launches programs with a special command
+        // line argument of the form -psn_XXX: just ignore it silently
+        // (and don't exit!)
         i++;
-#else
+      }
+#endif
+      else {
         fprintf(stderr, "Unknown option '%s'\n", argv[i]);
         CTX.terminal = 1;
         Print_Usage(argv[0]);
         exit(1);
-#endif
       }
-    }
 
+    }
     else {
-      if(*nbfiles < MAX_OPEN_FILES)
-        TheFileNameTab[(*nbfiles)++] = argv[i++];
-      else {
-        fprintf(stderr, ERROR_STR "Too many input files\n");
-        exit(1);
-      }
+      List_Add(CTX.files, &argv[i++]);
     }
 
   }
 
-  strncpy(CTX.filename, TheFileNameTab[0], 255);
+  if(!List_Nbr(CTX.files))
+    strncpy(CTX.filename, CTX.default_filename_fullpath, 255);
+  else
+    strncpy(CTX.filename, *(char**)List_Pointer(CTX.files, 0), 255);
 }
