@@ -1,6 +1,6 @@
-// $Id: CAD.cpp,v 1.75 2004/05/12 22:51:07 geuzaine Exp $
+// $Id: CAD.cpp,v 1.95.2.2 2006/05/13 22:13:49 geuzaine Exp $
 //
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -109,9 +109,13 @@ int compare2Lists(List_T * List1, List_T * List2,
 {
   int i, found;
 
-  if(List_Nbr(List1) != List_Nbr(List2))
-    return List_Nbr(List1) - List_Nbr(List2);
+  if(!List_Nbr(List1) && !List_Nbr(List2))
+    return 0;
 
+  if(!List_Nbr(List1) || !List_Nbr(List2) || 
+     (List_Nbr(List1) != List_Nbr(List2)))
+    return List_Nbr(List1) - List_Nbr(List2);
+  
   List_T *List1Prime = List_Create(List_Nbr(List1), 1, List1->size);
   List_T *List2Prime = List_Create(List_Nbr(List2), 1, List2->size);
   List_Copy(List1, List1Prime);
@@ -236,13 +240,22 @@ void CopyVertex(Vertex * v, Vertex * vv)
 
 Vertex *DuplicateVertex(Vertex * v)
 {
-  Vertex *pv;
-  pv = Create_Vertex(NEWPOINT(), 0, 0, 0, 0, 0);
+  if(!v) return NULL;
+  Vertex *pv = Create_Vertex(NEWPOINT(), 0, 0, 0, 0, 0);
   CopyVertex(v, pv);
   Tree_Insert(THEM->Points, &pv);
   return pv;
 }
 
+int compareAbsCurve(const void *a, const void *b)
+{
+  Curve **q, **w;
+
+  q = (Curve **) a;
+  w = (Curve **) b;
+  return (abs((*q)->Num) - abs((*w)->Num));
+}
+
 void CopyCurve(Curve * c, Curve * cc)
 {
   int i, j;
@@ -362,7 +375,7 @@ void CopyShape(int Type, int Num, int *New)
   switch (Type) {
   case MSH_POINT:
     if(!(v = FindPoint(Num, THEM))) {
-      Msg(GERROR, "Unknown Vertex %d", Num);
+      Msg(GERROR, "Unknown vertex %d", Num);
       return;
     }
     newv = DuplicateVertex(v);
@@ -373,11 +386,13 @@ void CopyShape(int Type, int Num, int *New)
   case MSH_SEGM_BSPLN:
   case MSH_SEGM_BEZIER:
   case MSH_SEGM_CIRC:
+  case MSH_SEGM_CIRC_INV:
   case MSH_SEGM_ELLI:
+  case MSH_SEGM_ELLI_INV:
   case MSH_SEGM_NURBS:
   case MSH_SEGM_PARAMETRIC:
     if(!(c = FindCurve(Num, THEM))) {
-      Msg(GERROR, "Unknown Curve %d", Num);
+      Msg(GERROR, "Unknown curve %d", Num);
       return;
     }
     newc = DuplicateCurve(c);
@@ -388,7 +403,7 @@ void CopyShape(int Type, int Num, int *New)
   case MSH_SURF_REGL:
   case MSH_SURF_PLAN:
     if(!(s = FindSurface(Num, THEM))) {
-      Msg(GERROR, "Unknown Surface %d", Num);
+      Msg(GERROR, "Unknown surface %d", Num);
       return;
     }
     news = DuplicateSurface(s);
@@ -400,8 +415,6 @@ void CopyShape(int Type, int Num, int *New)
   }
 }
 
-// FIXME: we still need to actually free the entity!
-
 void DeletePoint(int ip)
 {
   Vertex *v = FindPoint(ip, THEM);
@@ -412,14 +425,17 @@ void DeletePoint(int ip)
     Curve *c;
     List_Read(Curves, i, &c);
     for(int j = 0; j < List_Nbr(c->Control_Points); j++) {
-      if(!compareVertex(List_Pointer(c->Control_Points, j), &v))
+      if(!compareVertex(List_Pointer(c->Control_Points, j), &v)){
+	List_Delete(Curves);
         return;
+      }
     }
   }
   List_Delete(Curves);
   if(v->Num == THEM->MaxPointNum)
     THEM->MaxPointNum--;
   Tree_Suppress(THEM->Points, &v);
+  Free_Vertex(&v, NULL);
 }
 
 void DeleteCurve(int ip)
@@ -432,20 +448,21 @@ void DeleteCurve(int ip)
     Surface *s;
     List_Read(Surfs, i, &s);
     for(int j = 0; j < List_Nbr(s->Generatrices); j++) {
-      if(!compareCurve(List_Pointer(s->Generatrices, j), &c))
+      if(!compareAbsCurve(List_Pointer(s->Generatrices, j), &c)){
+	List_Delete(Surfs);
         return;
+      }
     }
   }
   List_Delete(Surfs);
   if(c->Num == THEM->MaxLineNum)
     THEM->MaxLineNum--;
   Tree_Suppress(THEM->Curves, &c);
+  Free_Curve(&c, NULL);
 }
 
 void DeleteSurface(int is)
 {
-  // Il faut absolument coder une
-  // structure coherente pour les volumes.
   Surface *s = FindSurface(is, THEM);
   if(!s)
     return;
@@ -454,14 +471,28 @@ void DeleteSurface(int is)
     Volume *v;
     List_Read(Vols, i, &v);
     for(int j = 0; j < List_Nbr(v->Surfaces); j++) {
-      if(!compareCurve(List_Pointer(v->Surfaces, j), &s))
+      if(!compareSurface(List_Pointer(v->Surfaces, j), &s)){
+	List_Delete(Vols);
         return;
+      }
     }
   }
   List_Delete(Vols);
   if(s->Num == THEM->MaxSurfaceNum)
     THEM->MaxSurfaceNum--;
   Tree_Suppress(THEM->Surfaces, &s);
+  Free_Surface(&s, NULL);
+}
+
+void DeleteVolume(int iv)
+{
+  Volume *v = FindVolume(iv, THEM);
+  if(!v)
+    return;
+  if(v->Num == THEM->MaxVolumeNum)
+    THEM->MaxVolumeNum--;
+  Tree_Suppress(THEM->Volumes, &v);
+  Free_Volume(&v, NULL);
 }
 
 void DeleteShape(int Type, int Num)
@@ -475,10 +506,13 @@ void DeleteShape(int Type, int Num)
   case MSH_SEGM_BSPLN:
   case MSH_SEGM_BEZIER:
   case MSH_SEGM_CIRC:
+  case MSH_SEGM_CIRC_INV:
   case MSH_SEGM_ELLI:
+  case MSH_SEGM_ELLI_INV:
   case MSH_SEGM_NURBS:
   case MSH_SEGM_PARAMETRIC:
     DeleteCurve(Num);
+    DeleteCurve(-Num);
     break;
   case MSH_SURF_NURBS:
   case MSH_SURF_TRIC:
@@ -486,6 +520,9 @@ void DeleteShape(int Type, int Num)
   case MSH_SURF_PLAN:
     DeleteSurface(Num);
     break;
+  case MSH_VOLUME:
+    DeleteVolume(Num);
+    break;
   default:
     Msg(GERROR, "Impossible to delete entity %d (of type %d)", Num, Type);
     break;
@@ -529,18 +566,23 @@ void ColorShape(int Type, int Num, unsigned int Color)
   case MSH_SEGM_BSPLN:
   case MSH_SEGM_BEZIER:
   case MSH_SEGM_CIRC:
+  case MSH_SEGM_CIRC_INV:
   case MSH_SEGM_ELLI:
+  case MSH_SEGM_ELLI_INV:
   case MSH_SEGM_NURBS:
   case MSH_SEGM_PARAMETRIC:
+  case MSH_SEGM_DISCRETE:
     ColorCurve(Num, Color);
     break;
   case MSH_SURF_NURBS:
   case MSH_SURF_TRIC:
   case MSH_SURF_REGL:
   case MSH_SURF_PLAN:
+  case MSH_SURF_DISCRETE:
     ColorSurface(Num, Color);
     break;
   case MSH_VOLUME:
+  case MSH_VOLUME_DISCRETE:
     ColorVolume(Num, Color);
     break;
   default:
@@ -559,18 +601,23 @@ void VisibilityShape(int Type, int Num, int Mode)
   case MSH_SEGM_BSPLN:
   case MSH_SEGM_BEZIER:
   case MSH_SEGM_CIRC:
+  case MSH_SEGM_CIRC_INV:
   case MSH_SEGM_ELLI:
+  case MSH_SEGM_ELLI_INV:
   case MSH_SEGM_NURBS:
   case MSH_SEGM_PARAMETRIC:
+  case MSH_SEGM_DISCRETE:
     SetVisibilityByNumber(Num, 3, Mode);
     break;
   case MSH_SURF_NURBS:
   case MSH_SURF_TRIC:
   case MSH_SURF_REGL:
   case MSH_SURF_PLAN:
+  case MSH_SURF_DISCRETE:
     SetVisibilityByNumber(Num, 4, Mode);
     break;
   case MSH_VOLUME:
+  case MSH_VOLUME_DISCRETE:
     SetVisibilityByNumber(Num, 5, Mode);
     break;
   default:
@@ -584,20 +631,23 @@ Curve *CreateReversedCurve(Mesh * M, Curve * c)
   Vertex *e1, *e2, *e3, *e4;
   int i;
   newc = Create_Curve(-c->Num, c->Typ, 1, NULL, NULL, -1, -1, 0., 1.);
-  newc->Control_Points =
-    List_Create(List_Nbr(c->Control_Points), 1, sizeof(Vertex *));
-  if(c->Typ == MSH_SEGM_ELLI || c->Typ == MSH_SEGM_ELLI_INV) {
-    List_Read(c->Control_Points, 0, &e1);
-    List_Read(c->Control_Points, 1, &e2);
-    List_Read(c->Control_Points, 2, &e3);
-    List_Read(c->Control_Points, 3, &e4);
-    List_Add(newc->Control_Points, &e4);
-    List_Add(newc->Control_Points, &e2);
-    List_Add(newc->Control_Points, &e3);
-    List_Add(newc->Control_Points, &e1);
+
+  if(List_Nbr(c->Control_Points)){
+    newc->Control_Points =
+      List_Create(List_Nbr(c->Control_Points), 1, sizeof(Vertex *));
+    if(c->Typ == MSH_SEGM_ELLI || c->Typ == MSH_SEGM_ELLI_INV) {
+      List_Read(c->Control_Points, 0, &e1);
+      List_Read(c->Control_Points, 1, &e2);
+      List_Read(c->Control_Points, 2, &e3);
+      List_Read(c->Control_Points, 3, &e4);
+      List_Add(newc->Control_Points, &e4);
+      List_Add(newc->Control_Points, &e2);
+      List_Add(newc->Control_Points, &e3);
+      List_Add(newc->Control_Points, &e1);
+    }
+    else
+      List_Invert(c->Control_Points, newc->Control_Points);
   }
-  else
-    List_Invert(c->Control_Points, newc->Control_Points);
 
   if(c->Typ == MSH_SEGM_NURBS && c->k) {
     newc->k =
@@ -615,7 +665,6 @@ Curve *CreateReversedCurve(Mesh * M, Curve * c)
     newc->Typ = MSH_SEGM_ELLI_INV;
   if(c->Typ == MSH_SEGM_ELLI_INV)
     newc->Typ = MSH_SEGM_ELLI;
-  newc->Vertices = List_Create(10, 1, sizeof(Vertex *));
   newc->Method = c->Method;
   newc->degre = c->degre;
   newc->beg = c->end;
@@ -853,7 +902,32 @@ static void vecmat4x4(double mat[4][4], double vec[4], double res[4])
   }
 }
 
-void ApplyTransformationToPoint(double matrix[4][4], Vertex * v)
+void printCurve(Curve * c)
+{
+  Vertex *v;
+  int N = List_Nbr(c->Control_Points);
+  Msg(DEBUG, "Curve %d %d cp (%d->%d)", c->Num, N, c->beg->Num, c->end->Num);
+  for(int i = 0; i < N; i++) {
+    List_Read(c->Control_Points, i, &v);
+    Msg(DEBUG, "Vertex %d (%g,%g,%g,%g)", v->Num, v->Pos.X, v->Pos.Y,
+        v->Pos.Z, v->lc);
+  }
+}
+
+void printSurface(Surface * s)
+{
+  Curve *c;
+  int N = List_Nbr(s->Generatrices);
+
+  Msg(DEBUG, "Surface %d, %d generatrices", s->Num, N);
+  for(int i = 0; i < N; i++) {
+    List_Read(s->Generatrices, i, &c);
+    printCurve(c);
+  }
+}
+
+void ApplyTransformationToPoint(double matrix[4][4], Vertex * v,
+				bool end_curve_surface=false)
 {
   double pos[4], vec[4];
 
@@ -875,12 +949,52 @@ void ApplyTransformationToPoint(double matrix[4][4], Vertex * v)
   v->Pos.Y = pos[1];
   v->Pos.Z = pos[2];
   v->w = pos[3];
+
+  // Warning: in theory we should always redo these checks if
+  // end_curve_surface is true; but in practice this is so slow for
+  // big models that we need to provide a way to bypass it (which is
+  // OK if the guy who builds the geometry knowns what he's
+  // doing). Instead of adding one more option, let's just bypass all
+  // the checks if auto_coherence==0...
+  if(CTX.geom.auto_coherence && end_curve_surface){
+    List_T *All = Tree2List(THEM->Curves);
+    for(int i = 0; i < List_Nbr(All); i++) {
+      Curve *c;
+      List_Read(All, i, &c);
+      for(int j = 0; j < List_Nbr(c->Control_Points); j++) {
+	Vertex *pv = *(Vertex **) List_Pointer(c->Control_Points, j);
+	if(pv->Num == v->Num){
+	  End_Curve(c);
+	  break;
+	}
+      }
+    }
+    List_Delete(All);
+    All = Tree2List(THEM->Surfaces);
+    for(int i = 0; i < List_Nbr(All); i++) {
+      Surface *s;
+      List_Read(All, i, &s);
+      for(int j = 0; j < List_Nbr(s->Control_Points); j++) {
+	Vertex *pv = *(Vertex **) List_Pointer(s->Control_Points, j);
+	if(pv->Num == v->Num){
+	  End_Surface(s);
+	  break;
+	}
+      }
+    }
+    List_Delete(All);
+  }
 }
 
 void ApplyTransformationToCurve(double matrix[4][4], Curve * c)
 {
   Vertex *v;
 
+  if(!c->beg || !c->end){
+    Msg(GERROR, "Cannot transform curve with no begin/end points");
+    return;
+  }
+
   ApplyTransformationToPoint(matrix, c->beg);
   ApplyTransformationToPoint(matrix, c->end);
 
@@ -891,18 +1005,6 @@ void ApplyTransformationToCurve(double matrix[4][4], Curve * c)
   End_Curve(c);
 }
 
-void printCurve(Curve * c)
-{
-  Vertex *v;
-  int N = List_Nbr(c->Control_Points);
-  Msg(DEBUG, "Curve %d %d cp (%d->%d)", c->Num, N, c->beg->Num, c->end->Num);
-  for(int i = 0; i < N; i++) {
-    List_Read(c->Control_Points, i, &v);
-    Msg(DEBUG, "Vertex %d (%g,%g,%g,%g)", v->Num, v->Pos.X, v->Pos.Y,
-        v->Pos.Z, v->lc);
-  }
-}
-
 void ApplyTransformationToSurface(double matrix[4][4], Surface * s)
 {
   Curve *c;
@@ -911,7 +1013,11 @@ void ApplyTransformationToSurface(double matrix[4][4], Surface * s)
 
   for(i = 0; i < List_Nbr(s->Generatrices); i++) {
     List_Read(s->Generatrices, i, &c);
-    ApplyTransformationToCurve(matrix, c);
+    // FIXME: this fixes benchmarks/bug/transfo_neg_curves.geo, but is
+    // it the correct fix?
+    //ApplyTransformationToCurve(matrix, c);
+    Curve *cc = FindCurve(abs(c->Num), THEM);
+    ApplyTransformationToCurve(matrix, cc);
   }
   for(i = 0; i < List_Nbr(s->Control_Points); i++) {
     List_Read(s->Control_Points, i, &v);
@@ -920,18 +1026,6 @@ void ApplyTransformationToSurface(double matrix[4][4], Surface * s)
   End_Surface(s);
 }
 
-void printSurface(Surface * s)
-{
-  Curve *c;
-  int N = List_Nbr(s->Generatrices);
-
-  Msg(DEBUG, "Surface %d, %d generatrices", s->Num, N);
-  for(int i = 0; i < N; i++) {
-    List_Read(s->Generatrices, i, &c);
-    printCurve(c);
-  }
-}
-
 void ApplicationOnShapes(double matrix[4][4], List_T * ListShapes)
 {
   int i;
@@ -948,23 +1042,25 @@ void ApplicationOnShapes(double matrix[4][4], List_T * ListShapes)
     case MSH_POINT:
       v = FindPoint(O.Num, THEM);
       if(v)
-        ApplyTransformationToPoint(matrix, v);
+        ApplyTransformationToPoint(matrix, v, true);
       else
-        Msg(GERROR, "Unknown Point %d", O.Num);
+        Msg(GERROR, "Unknown point %d", O.Num);
       break;
     case MSH_SEGM_LINE:
     case MSH_SEGM_SPLN:
     case MSH_SEGM_BSPLN:
     case MSH_SEGM_BEZIER:
     case MSH_SEGM_CIRC:
+    case MSH_SEGM_CIRC_INV:
     case MSH_SEGM_ELLI:
+    case MSH_SEGM_ELLI_INV:
     case MSH_SEGM_NURBS:
     case MSH_SEGM_PARAMETRIC:
       c = FindCurve(O.Num, THEM);
       if(c)
         ApplyTransformationToCurve(matrix, c);
       else
-        Msg(GERROR, "Unknown Curve %d", O.Num);
+        Msg(GERROR, "Unknown curve %d", O.Num);
       break;
     case MSH_SURF_NURBS:
     case MSH_SURF_REGL:
@@ -974,7 +1070,7 @@ void ApplicationOnShapes(double matrix[4][4], List_T * ListShapes)
       if(s)
         ApplyTransformationToSurface(matrix, s);
       else
-        Msg(GERROR, "Unknown Surface %d", O.Num);
+        Msg(GERROR, "Unknown surface %d", O.Num);
       break;
     default:
       Msg(GERROR, "Impossible to transform entity %d (of type %d)", O.Num,
@@ -1016,7 +1112,6 @@ void DilatShapes(double X, double Y, double Z, double A,
     ReplaceAllDuplicates(THEM);
 }
 
-
 void RotateShapes(double Ax, double Ay, double Az,
                   double Px, double Py, double Pz,
                   double alpha, List_T * ListShapes, int final)
@@ -1278,6 +1373,11 @@ int Extrude_ProtudeCurve(int type, int ic,
     return 0;
   }
 
+  if(!pc->beg || !pc->end){
+    Msg(GERROR, "Cannot extrude curve with no begin/end points");
+    return 0;
+  }
+
   Msg(DEBUG, "Extrude Curve %d", ic);
 
   chapeau = DuplicateCurve(pc);
@@ -1413,13 +1513,14 @@ int Extrude_ProtudeSurface(int type, int is,
 			   double T0, double T1, double T2,
 			   double A0, double A1, double A2,
 			   double X0, double X1, double X2, double alpha,
-			   int NewVolume, ExtrudeParams * e)
+			   Volume **pv, ExtrudeParams * e)
 {
   double matrix[4][4], T[3], Ax[3];
   Curve *c, *c2;
   int i;
   Surface *s, *ps, *chapeau;
-  Volume *pv = NULL;
+
+  *pv = NULL;
 
   if(!(ps = FindSurface(is, THEM)))
     return 0;
@@ -1439,7 +1540,7 @@ int Extrude_ProtudeSurface(int type, int is,
     List_Read(chapeau->Generatrices, i, &c);
     if(c->Num < 0)
       if(!(c = FindCurve(-c->Num, THEM))) {
-        Msg(GERROR, "Unknown Curve %d", -c->Num);
+        Msg(GERROR, "Unknown curve %d", -c->Num);
         return ps->Num;
       }
     c->Extrude = new ExtrudeParams(COPIED_ENTITY);
@@ -1450,25 +1551,44 @@ int Extrude_ProtudeSurface(int type, int is,
       c->Extrude->mesh = e->mesh;
   }
 
-  if(NewVolume) {
-    pv = Create_Volume(NewVolume, 0);
-    pv->Extrude = new ExtrudeParams;
-    pv->Extrude->fill(type, T0, T1, T2, A0, A1, A2, X0, X1, X2, alpha);
-    pv->Extrude->geo.Source = is;
-    if(e)
-      pv->Extrude->mesh = e->mesh;
-    if(pv)
-      List_Add(pv->Surfaces, &ps);
-  }
-  if(pv)
-    List_Add(pv->Surfaces, &chapeau);
+  // FIXME: this is a really ugly hack for backward compatibility, so
+  // that we don't screw up the old .geo files too much. (Before
+  // version 1.54, we didn't always create new volumes during "Extrude
+  // Surface". Now we do, but with "CTX.geom.old_newreg==1", this
+  // bumps the NEWREG() counter, and thus changes the whole automatic
+  // numbering sequence.) So we locally force old_newreg to 0: in most
+  // cases, since we define points, curves, etc., before defining
+  // volumes, the NEWVOLUME() call below will return a fairly low
+  // number, that will not interfere with the other numbers...
+  int tmp = CTX.geom.old_newreg;
+  CTX.geom.old_newreg = 0;
+  Volume *v = Create_Volume(NEWVOLUME(), MSH_VOLUME);
+  CTX.geom.old_newreg = tmp;
+
+  v->Extrude = new ExtrudeParams;
+  v->Extrude->fill(type, T0, T1, T2, A0, A1, A2, X0, X1, X2, alpha);
+  v->Extrude->geo.Source = is;
+  if(e)
+    v->Extrude->mesh = e->mesh;
+  int ori = -1;
+  List_Add(v->Surfaces, &ps);
+  List_Add(v->SurfacesOrientations, &ori);
+  ori = 1;
+  List_Add(v->Surfaces, &chapeau);
+  List_Add(v->SurfacesOrientations, &ori);
 
   for(i = 0; i < List_Nbr(ps->Generatrices); i++) {
     List_Read(ps->Generatrices, i, &c);
     Extrude_ProtudeCurve(type, c->Num, T0, T1, T2, A0, A1, A2, X0, X1, X2,
 			 alpha, &s, 0, e);
-    if(pv && s)
-      List_Add(pv->Surfaces, &s);
+    if(s){
+      if(c < 0)
+	ori = -1;
+      else 
+	ori = 1;
+      List_Add(v->Surfaces, &s);
+      List_Add(v->SurfacesOrientations, &ori);
+    }
   }
 
   switch (type) {
@@ -1536,14 +1656,15 @@ int Extrude_ProtudeSurface(int type, int is,
     return ps->Num;
   }
 
-  // why do we do this? only for backward compatibility?
+  // FIXME: why do we do this? only for backward compatibility?
   Tree_Suppress(THEM->Surfaces, &chapeau);
   chapeau->Num = NEWSURFACE();
   THEM->MaxSurfaceNum = chapeau->Num;
   Tree_Add(THEM->Surfaces, &chapeau);
 
-  if(pv)
-    Tree_Add(THEM->Volumes, &pv);
+  Tree_Add(THEM->Volumes, &v);
+
+  *pv = v;
 
   if(CTX.geom.auto_coherence)
     ReplaceAllDuplicates(THEM);
@@ -1553,25 +1674,142 @@ int Extrude_ProtudeSurface(int type, int is,
   return chapeau->Num;
 }
 
+void ExtrudeShape(int extrude_type, int shape_type, int shape_num,
+		  double T0, double T1, double T2,
+		  double A0, double A1, double A2,
+		  double X0, double X1, double X2, double alpha,
+		  ExtrudeParams *e,
+		  List_T *out)
+{
+  Shape shape;
+  shape.Type = shape_type;
+  shape.Num = shape_num;
+  List_T *tmp = List_Create(1, 1, sizeof(Shape));
+  List_Add(tmp, &shape);
+  ExtrudeShapes(extrude_type, tmp,
+		T0, T1, T2,
+		A0, A1, A2,
+		X0, X1, X2, alpha,
+		e,
+		out);
+  List_Delete(tmp);
+}
+
+void ExtrudeShapes(int type, List_T *in, 
+		   double T0, double T1, double T2,
+		   double A0, double A1, double A2,
+		   double X0, double X1, double X2, double alpha,
+		   ExtrudeParams *e,
+		   List_T *out)
+{
+  Shape O, TheShape;
+  Curve *pc, *prc;
+  Surface *ps;
+  Volume *pv;
+      
+  for(int i = 0; i < List_Nbr(in); i++){
+    List_Read(in, i, &O);
+    switch(O.Type){
+    case MSH_POINT:
+      TheShape.Num = Extrude_ProtudePoint(type, O.Num, T0, T1, T2,
+					  A0, A1, A2, X0, X1, X2, alpha,
+					  &pc, &prc, 1, e);
+      TheShape.Type = MSH_POINT;
+      List_Add(out, &TheShape);
+      if(pc){
+	TheShape.Num = pc->Num;
+	TheShape.Type = pc->Typ;
+	List_Add(out, &TheShape);
+      }
+      break;
+    case MSH_SEGM_LINE:
+    case MSH_SEGM_SPLN:
+    case MSH_SEGM_BSPLN:
+    case MSH_SEGM_BEZIER:
+    case MSH_SEGM_CIRC:
+    case MSH_SEGM_CIRC_INV:
+    case MSH_SEGM_ELLI:
+    case MSH_SEGM_ELLI_INV:
+    case MSH_SEGM_NURBS:
+    case MSH_SEGM_PARAMETRIC:
+      TheShape.Num = Extrude_ProtudeCurve(type, O.Num, T0, T1, T2,
+					  A0, A1, A2, X0, X1, X2, alpha,
+					  &ps, 1, e);
+      pc = FindCurve(TheShape.Num, THEM);
+      if(!pc){
+	//Msg(WARNING, "Unknown curve %d", TheShape.Num);
+	TheShape.Type = 0;
+      }
+      else{
+	TheShape.Type = pc->Typ;
+      }
+      List_Add(out, &TheShape);
+      if(ps){
+	TheShape.Num = ps->Num;
+	TheShape.Type = ps->Typ;
+	List_Add(out, &TheShape);
+      }
+      break;
+    case MSH_SURF_NURBS:
+    case MSH_SURF_REGL:
+    case MSH_SURF_TRIC:
+    case MSH_SURF_PLAN:
+      TheShape.Num = Extrude_ProtudeSurface(type, O.Num, T0, T1, T2,
+					    A0, A1, A2, X0, X1, X2, alpha,
+					    &pv, e);
+      ps = FindSurface(TheShape.Num, THEM);
+      if(!ps){
+	//Msg(WARNING, "Unknown surface %d", TheShape.Num);
+	TheShape.Type = 0;
+      }
+      else{
+	TheShape.Type = ps->Typ;
+      }
+      List_Add(out, &TheShape);
+      if(pv){
+	TheShape.Num = pv->Num;
+	TheShape.Type = pv->Typ;
+	List_Add(out, &TheShape);
+      }
+      break;
+    default:
+      Msg(GERROR, "Impossible to extrude entity %d (of type %d)", O.Num,
+          O.Type);
+      break;
+    }
+  }
+}
+
 // Duplicate removal
 
 int compareTwoCurves(const void *a, const void *b)
 {
-  Curve *c1, *c2;
-  c1 = *(Curve **) a;
-  c2 = *(Curve **) b;
+  Curve *c1 = *(Curve **) a;
+  Curve *c2 = *(Curve **) b;
   int comp;
 
-  if(c1->Typ != c2->Typ)
-    return c1->Typ - c2->Typ;
-  
+  if(c1->Typ != c2->Typ){
+    if((c1->Typ == MSH_SEGM_CIRC && c2->Typ == MSH_SEGM_CIRC_INV) ||
+       (c1->Typ == MSH_SEGM_CIRC_INV && c2->Typ == MSH_SEGM_CIRC) ||
+       (c1->Typ == MSH_SEGM_ELLI && c2->Typ == MSH_SEGM_ELLI_INV) ||
+       (c1->Typ == MSH_SEGM_ELLI_INV && c2->Typ == MSH_SEGM_ELLI)){
+      // this is still ok
+    }
+    else
+      return c1->Typ - c2->Typ;
+  }
+
   if(List_Nbr(c1->Control_Points) != List_Nbr(c2->Control_Points))
     return List_Nbr(c1->Control_Points) - List_Nbr(c2->Control_Points);
   
   if(!List_Nbr(c1->Control_Points)){
+    if(!c1->beg || !c2->beg)
+      return 1;
     comp = compareVertex(&c1->beg, &c2->beg);
     if(comp)
       return comp;
+    if(!c1->end || !c2->end)
+      return 1;
     comp = compareVertex(&c1->end, &c2->end);
     if(comp)
       return comp;
@@ -1590,20 +1828,10 @@ int compareTwoCurves(const void *a, const void *b)
   return 0;
 }
 
-int compareAbsCurve(const void *a, const void *b)
-{
-  Curve **q, **w;
-
-  q = (Curve **) a;
-  w = (Curve **) b;
-  return (abs((*q)->Num) - abs((*w)->Num));
-}
-
 int compareTwoSurfaces(const void *a, const void *b)
 {
-  Surface *s1, *s2;
-  s1 = *(Surface **) a;
-  s2 = *(Surface **) b;
+  Surface *s1 = *(Surface **) a;
+  Surface *s2 = *(Surface **) b;
   return compare2Lists(s1->Generatrices, s2->Generatrices, compareAbsCurve);
 }
 
@@ -1758,7 +1986,7 @@ void ReplaceDuplicateCurves(Mesh * m)
       if(!Tree_Search(allNonDuplicatedCurves, &c)) {
         Tree_Insert(allNonDuplicatedCurves, &c);
         if(!(c2 = FindCurve(-c->Num, m))) {
-          Msg(GERROR, "Unknown Curve %d", -c->Num);
+          Msg(GERROR, "Unknown curve %d", -c->Num);
           List_Delete(All);
           return;
         }
@@ -1767,7 +1995,7 @@ void ReplaceDuplicateCurves(Mesh * m)
       else {
         Tree_Suppress(m->Curves, &c);
         if(!(c2 = FindCurve(-c->Num, m))) {
-          Msg(GERROR, "Unknown Curve %d", -c->Num);
+          Msg(GERROR, "Unknown curve %d", -c->Num);
           List_Delete(All);
           return;
         }
@@ -1879,19 +2107,7 @@ void ReplaceAllDuplicates(Mesh * m)
 }
 
 
-// Inetersection routines
-// FIXME: this code is full of crap :-)
-
-/*
-  S = (sx(u,v),sy(u,v),sz(u,v))
-  C = (cx(w),cy(w),cz(w))
-
-  sx - cx = 0
-  sy - cy = 0
-  sz - cz = 0
-  
-  3eqs 3incs
-*/
+// Projection of point on curve or surface
 
 static Curve *CURVE, *CURVE_2;
 static Surface *SURFACE;
@@ -2128,12 +2344,6 @@ bool ProjectPointOnSurface(Surface * s, Vertex * p, double *u, double *v)
     deb = 0;
   }
 
-  if(p->Num == 160) {
-    lmin += 1.e90;
-    if(p->Num == 133)
-      UMAX = s->ku[s->Nu + s->OrderU];
-  }
-
   if(try_a_value(SURFACE, VERTEX, SURFACE->ku[0] + VERTEX->u,
                  SURFACE->kv[0], u, v))
     return true;
@@ -2196,6 +2406,14 @@ bool ProjectPointOnSurface(Surface * s, Vertex * p, double *u, double *v)
   return true;
 }
 
+// Intersection of curves/surfaces (FIXME: this code is full of crap)
+
+// S = (sx(u,v),sy(u,v),sz(u,v))  C = (cx(w),cy(w),cz(w))
+// sx - cx = 0
+// sy - cy = 0
+// sz - cz = 0
+// 3eqs 3unk
+
 bool IntersectCurveSurface(Curve * c, Surface * s)
 {
   double x[4];
@@ -2208,7 +2426,6 @@ bool IntersectCurveSurface(Curve * c, Surface * s)
   return true;
 }
 
-
 void DivideCurve(Curve * c, double u, Vertex * v, Curve ** c1, Curve ** c2)
 {
   (*c1) = Create_Curve(NEWLINE(), c->Typ, 1, NULL, NULL, -1, -1, 0., 1.);
