@@ -1,6 +1,6 @@
-// $Id: Geom.cpp,v 1.65 2004/05/31 18:36:20 geuzaine Exp $
+// $Id: Geom.cpp,v 1.97 2006/01/10 03:58:31 geuzaine Exp $
 //
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 // 
 // Please report all bugs and problems to <gmsh@geuz.org>.
 
+#include "BDS.h"
 #include "Gmsh.h"
 #include "GmshUI.h"
 #include "Numeric.h"
@@ -29,7 +30,8 @@
 #include "Draw.h"
 #include "Context.h"
 #include "Interpolation.h"
-#include "STL.h"
+#include "Plugin.h"
+#include "PluginManager.h"
 #include "gl2ps.h"
 
 extern Context_T CTX;
@@ -47,7 +49,7 @@ void Draw_Geo_Point(void *a, void *b)
     return;
 
   if(CTX.render_mode == GMSH_SELECT) {
-    glLoadName(0);
+    glPushName(0);
     glPushName(v->Num);
   }
 
@@ -63,32 +65,43 @@ void Draw_Geo_Point(void *a, void *b)
   }
 
   if(CTX.geom.points) {
-    if(CTX.geom.point_type) {
+    if(CTX.geom.point_type == 1) {
       if(v->Frozen)
-        Draw_Sphere(CTX.geom.point_sel_size, v->Pos.X, v->Pos.Y, v->Pos.Z, 
+	Draw_Sphere(CTX.geom.point_sel_size, v->Pos.X, v->Pos.Y, v->Pos.Z, 
 		    CTX.geom.light);
       else
-        Draw_Sphere(CTX.geom.point_size, v->Pos.X, v->Pos.Y, v->Pos.Z,
+	Draw_Sphere(CTX.geom.point_size, v->Pos.X, v->Pos.Y, v->Pos.Z,
 		    CTX.geom.light);
     }
+    else if(CTX.geom.point_type == 2) {
+      GMSH_Solve_Plugin *sp = GMSH_PluginManager::instance()->findSolverPlugin();
+      if(sp) {
+	sp-> GL_enhancePoint (v);
+      }
+      glBegin(GL_POINTS);
+      glVertex3d(v->Pos.X, v->Pos.Y, v->Pos.Z);
+      glEnd();
+    }
     else {
       glBegin(GL_POINTS);
       glVertex3d(v->Pos.X, v->Pos.Y, v->Pos.Z);
       glEnd();
     }
+
   }
 
   if(CTX.geom.points_num) {
     sprintf(Num, "%d", v->Num);
-    double offset = 0.5 * (CTX.geom.point_size + CTX.gl_fontsize) * CTX.pixel_equiv_x;
+    double offset = (0.5 * CTX.geom.point_size + 0.3 * CTX.gl_fontsize) * CTX.pixel_equiv_x;
     glRasterPos3d(v->Pos.X + offset / CTX.s[0],
-                  v->Pos.Y + offset / CTX.s[1],
-                  v->Pos.Z + offset / CTX.s[2]);
+		  v->Pos.Y + offset / CTX.s[1],
+		  v->Pos.Z + offset / CTX.s[2]);
     Draw_String(Num);
   }
 
   if(CTX.render_mode == GMSH_SELECT) {
     glPopName();
+    glPopName();
   }
 }
 
@@ -103,15 +116,15 @@ void Draw_Curve(void *a, void *b)
 
   Curve *c = *(Curve **) a;
 
-  if(c->Num < 0 || !(c->Visible & VIS_GEOM) || c->Dirty)
+  if(c->Num < 0 || !(c->Visible & VIS_GEOM))
     return;
 
   if(CTX.render_mode == GMSH_SELECT) {
-    glLoadName(1);
+    glPushName(1);
     glPushName(c->Num);
   }
 
-  if(c->ipar[3]) {
+  if(c->ipar[3] > 0) {
     glLineWidth(CTX.geom.line_sel_width);
     gl2psLineWidth(CTX.geom.line_sel_width * CTX.print.eps_line_width_factor);
     glColor4ubv((GLubyte *) & CTX.color.geom.line_sel);
@@ -138,40 +151,68 @@ void Draw_Curve(void *a, void *b)
       N = 10 * n;
       break;
     }
-    if(c->Typ == MSH_SEGM_DISCRETE) {
-      Simplex *s;
-      List_T *temp = Tree2List(c->Simplexes);
-      for(int i = 0; i < List_Nbr(temp); i++) {
-        List_Read(temp, i, &s);
-        glBegin(GL_LINE_STRIP);
-        glVertex3d(s->V[0]->Pos.X, s->V[0]->Pos.Y, s->V[0]->Pos.Z);
-        glVertex3d(s->V[1]->Pos.X, s->V[1]->Pos.Y, s->V[1]->Pos.Z);
-        glEnd();
+    if(c->Typ == MSH_SEGM_DISCRETE && THEM->bds) {
+      BDS_GeomEntity *g = THEM->bds->get_geom ( c->Num,1);	
+      std::list<BDS_Edge*>::iterator it  = g->e.begin();
+      std::list<BDS_Edge*>::iterator ite = g->e.end();
+      while (it!=ite){
+	BDS_Edge *e = (*it);
+	if(CTX.geom.line_type < 1) {
+	  glBegin(GL_LINES);
+	  glVertex3d(e->p1->X,e->p1->Y,e->p1->Z);
+	  glVertex3d(e->p2->X,e->p2->Y,e->p2->Z);
+	  glEnd();
+	}
+	else{
+	  x[0] = e->p1->X;
+	  y[0] = e->p1->Y;
+	  z[0] = e->p1->Z;
+	  x[1] = e->p2->X;
+	  y[1] = e->p2->Y;
+	  z[1] = e->p2->Z;
+	  Draw_Cylinder(c->ipar[3] > 0 ? CTX.geom.line_sel_width : CTX.geom.line_width,
+			x, y, z, CTX.geom.light);
+	}
+	++it;
       }
-      List_Delete(temp);
+    }
+    else if(c->Typ == MSH_SEGM_DISCRETE) {
+      // do nothing: we draw the elements in the mesh drawing routines
     }
     else {
-      if(CTX.geom.line_type) {
-        for(int i = 0; i < N - 1; i++) {
-          v = InterpolateCurve(c, (double)i / (double)(N - 1), 0);
-          dv = InterpolateCurve(c, (double)(i + 1) / (double)(N - 1), 0);
-          x[0] = v.Pos.X;
-          y[0] = v.Pos.Y;
-          z[0] = v.Pos.Z;
-          x[1] = dv.Pos.X;
-          y[1] = dv.Pos.Y;
-          z[1] = dv.Pos.Z;
-          Draw_Cylinder(c->ipar[3] ? CTX.geom.line_sel_width : CTX.geom.line_width,
+      if(CTX.geom.line_type >= 1) {
+	for(int i = 0; i < N - 1; i++) {
+	  v = InterpolateCurve(c, (double)i / (double)(N - 1), 0);
+	  dv = InterpolateCurve(c, (double)(i + 1) / (double)(N - 1), 0);
+	  x[0] = v.Pos.X;
+	  y[0] = v.Pos.Y;
+	  z[0] = v.Pos.Z;
+	  x[1] = dv.Pos.X;
+	  y[1] = dv.Pos.Y;
+	  z[1] = dv.Pos.Z;
+	  Draw_Cylinder(c->ipar[3] > 0 ? CTX.geom.line_sel_width : CTX.geom.line_width,
 			x, y, z, CTX.geom.light);
-        }
+	}
+	if(CTX.geom.line_type == 2) {
+	  GMSH_Solve_Plugin *sp = GMSH_PluginManager::instance()->findSolverPlugin();
+	  if(sp) {
+	    int NN=(N>1)?N:1;
+	    const double eps=0.e-2;
+	    for(int i = 0; i < NN - 1; i++) {
+	      v = InterpolateCurve(c, (double)i / (double)(NN - 1)-eps, 0);
+	      dv = InterpolateCurve(c, (double)(i + 1) / (double)(NN - 1)+eps, 0);
+	      sp-> GL_enhanceLine (c->Num,&v,&dv);
+	    }
+	  }
+	}
       }
       else {
-        glBegin(GL_LINE_STRIP);
-        for(int i = 0; i < N; i++) {
-          v = InterpolateCurve(c, (double)i / (double)(N - 1), 0);
-          glVertex3d(v.Pos.X, v.Pos.Y, v.Pos.Z);
-        }
-        glEnd();
+	glBegin(GL_LINE_STRIP);
+	for(int i = 0; i < N; i++) {
+	  v = InterpolateCurve(c, (double)i / (double)(N - 1), 0);
+	  glVertex3d(v.Pos.X, v.Pos.Y, v.Pos.Z);
+	}
+	glEnd();
       }
     }
   }
@@ -179,10 +220,10 @@ void Draw_Curve(void *a, void *b)
   if(CTX.geom.lines_num) {
     v = InterpolateCurve(c, 0.5, 0);
     sprintf(Num, "%d", c->Num);
-    double offset = 0.5 * (CTX.geom.line_width + CTX.gl_fontsize) * CTX.pixel_equiv_x;
+    double offset = (0.5 * CTX.geom.line_width + 0.3 * CTX.gl_fontsize) * CTX.pixel_equiv_x;
     glRasterPos3d(v.Pos.X + offset / CTX.s[0],
-                  v.Pos.Y + offset / CTX.s[1],
-                  v.Pos.Z + offset / CTX.s[2]);
+		  v.Pos.Y + offset / CTX.s[1],
+		  v.Pos.Z + offset / CTX.s[2]);
     Draw_String(Num);
   }
 
@@ -197,17 +238,18 @@ void Draw_Curve(void *a, void *b)
     Draw_Vector(CTX.vector_type, 0, CTX.arrow_rel_head_radius, 
 		CTX.arrow_rel_stem_length, CTX.arrow_rel_stem_radius,
 		v.Pos.X, v.Pos.Y, v.Pos.Z,
-                dv.Pos.X, dv.Pos.Y, dv.Pos.Z, NULL, CTX.geom.light);
+		dv.Pos.X, dv.Pos.Y, dv.Pos.Z, CTX.geom.light);
   }
 
   if(CTX.render_mode == GMSH_SELECT) {
     glPopName();
+    glPopName();
   }
 }
 
 // Surfaces
 
-void put_Z(Vertex * v, Surface * s)
+void putZ(Vertex * v, Surface * s)
 {
   Vertex V;
   V.Pos.X = s->a;
@@ -255,31 +297,77 @@ int isPointOnPlanarSurface(Surface * S, double X, double Y, double Z,
   return 0;
 }
 
-void Draw_Triangulated_Surface(Surface * s)
+void getPlaneSurfaceNormal(Surface *s, double x, double y, double z, double n[3])
 {
-  int k = 0;
-  double *points, *p1, *p2, *p3;
+  double t1[3], t2[3];
+  Curve *c;
 
-  if(CTX.geom.surfaces) {
-    if(CTX.geom.light) glEnable(GL_LIGHTING);
-    glEnable(GL_POLYGON_OFFSET_FILL);
-    glBegin(GL_TRIANGLES);
-    while (k < List_Nbr(s->thePolyRep->polygons)){
-      points = (double*)List_Pointer(s->thePolyRep->polygons,k);
-      k+= ((int)points[0] + 1);
-      if (points[0] == 3){
-	p1 = (double*)List_Pointer(s->thePolyRep->points_and_normals, 6*(int)points[1]);
-	p2 = (double*)List_Pointer(s->thePolyRep->points_and_normals, 6*(int)points[2]);
-	p3 = (double*)List_Pointer(s->thePolyRep->points_and_normals, 6*(int)points[3]);
-	if(CTX.geom.light) glNormal3dv(&p1[3]);
-	glVertex3d(p1[0],p1[1],p1[2]);
-	if(CTX.geom.light) glNormal3dv(&p2[3]);
-	glVertex3d(p2[0],p2[1],p2[2]);
-	if(CTX.geom.light) glNormal3dv(&p3[3]);
-	glVertex3d(p3[0],p3[1],p3[2]);
+  for(int i = 0; i < 3; i++)
+    n[i] = 0.0;
+
+  if(s->Typ != MSH_SURF_PLAN)
+    return;
+
+  // We cannot use s->plan or InterpolateSurface here (they both rely
+  // on the mean plane, which borks the orientation). So we need to
+  // use this trick:
+
+  if(List_Nbr(s->Generatrices) > 1){
+    List_Read(s->Generatrices, 0, &c);
+    if(c->beg){
+      t1[0] = x - c->beg->Pos.X;
+      t1[1] = y - c->beg->Pos.Y;
+      t1[2] = z - c->beg->Pos.Z;
+      // 1) try to get a point close to 'beg' on the same curve
+      // 2) if we are really unlucky and these two points are aligned 
+      // with (x,y,z), which we know is inside or on the boundary of
+      // the surface, then get a point from the next generatrice
+      // 3) repeat
+      for(int i = 0; i < List_Nbr(s->Generatrices); i++){
+	List_Read(s->Generatrices, i, &c);
+	Vertex v = InterpolateCurve(c, 0.1, 0);
+	t2[0] = x - v.Pos.X;
+	t2[1] = y - v.Pos.Y;
+	t2[2] = z - v.Pos.Z;
+	prodve(t1, t2, n);
+	if(norme(n))
+	  break;
       }
     }
-    glEnd();
+  }
+  if(List_Nbr(s->Generatrices) < 2 || !norme(n)){
+    Msg(WARNING, "Reverting to mean plane normal for surface %d", s->Num);
+    n[0] = s->a;
+    n[1] = s->b;
+    n[2] = s->c;
+    norme(n);
+  }
+}
+
+void Draw_Polygonal_Surface(Surface * s)
+{
+  if(CTX.geom.surfaces) {
+    if(CTX.geom.light) glEnable(GL_LIGHTING);
+    glEnable(GL_POLYGON_OFFSET_FILL); // always!
+    BDS_GeomEntity *g = THEM->bds->get_geom ( s->Num,2);	
+    std::list<BDS_Triangle*>::iterator it  = g->t.begin();
+    std::list<BDS_Triangle*>::iterator ite = g->t.end();
+    while (it!=ite) {
+      glBegin(GL_TRIANGLES);
+      BDS_Point *n[3];
+      BDS_Triangle *t = (*it);
+      double c[3];
+      t->getNodes (n);
+      normal_triangle (n[0],n[1],n[2],c);
+      if(CTX.geom.light) glNormal3dv(c);
+      glVertex3d(n[0]->X,n[0]->Y,n[0]->Z);
+      if(CTX.geom.light) glNormal3dv(c);
+      glVertex3d(n[1]->X,n[1]->Y,n[1]->Z);
+      if(CTX.geom.light) glNormal3dv(c);
+      glVertex3d(n[2]->X,n[2]->Y,n[2]->Z);
+      glEnd();
+      ++it;
+    }
     glDisable(GL_POLYGON_OFFSET_FILL);
     glDisable(GL_LIGHTING);
   }  
@@ -287,45 +375,42 @@ void Draw_Triangulated_Surface(Surface * s)
 
 void Draw_Plane_Surface(Surface * s)
 {
-  int i, j, k;
   Curve *c;
-  double minx = 0., miny = 0., maxx = 0., maxy = 0., t, n[3];
-  Vertex P1, P2, P3, V[4], vv, vv1, vv2;
-  char Num[100];
-  const int numPoints = 100;
-
-  if (s->thePolyRep) {
-    Draw_Triangulated_Surface(s);
-    return;
-  }
+  Vertex V[4], vv, vv1, vv2;
+  double n[3];
 
   if(!CTX.threads_lock && List_Nbr(s->Orientations) < 1) {
     CTX.threads_lock = 1;
 
     List_T *points = List_Create(10, 10, sizeof(Vertex *));
-    for(i = 0; i < List_Nbr(s->Generatrices); i++) {
+    for(int i = 0; i < List_Nbr(s->Generatrices); i++) {
       List_Read(s->Generatrices, i, &c);
-      for(j = 0; j < List_Nbr(c->Control_Points); j++) {
-        List_Add(points, List_Pointer(c->Control_Points, j));
+      for(int j = 0; j < List_Nbr(c->Control_Points); j++) {
+	List_Add(points, List_Pointer(c->Control_Points, j));
       }
     }
+
+    if(!List_Nbr(points)){
+      List_Delete(points);      
+      CTX.threads_lock = 0;
+      return;
+    }
     MeanPlane(points, s);
     List_Delete(points);
-
-    k = 0;
-
-    for(i = 0; i < List_Nbr(s->Generatrices); i++) {
+    
+    // compute (rough) bounding box of surface
+    double minx = 0., miny = 0., maxx = 0., maxy = 0.;
+    for(int i = 0; i < List_Nbr(s->Generatrices); i++) {
       List_Read(s->Generatrices, i, &c);
-      P1 = InterpolateCurve(c, 0.0, 0);
-      P2 = InterpolateCurve(c, 0.5, 0);
-      P3 = InterpolateCurve(c, 1.0, 0);
+      Vertex P1 = InterpolateCurve(c, 0.0, 0);
+      Vertex P2 = InterpolateCurve(c, 0.5, 0);
+      Vertex P3 = InterpolateCurve(c, 1.0, 0);
       Projette(&P1, s->plan);
       Projette(&P2, s->plan);
       Projette(&P3, s->plan);
-      if(!k) {
-        k = 1;
-        minx = maxx = P1.Pos.X;
-        miny = maxy = P1.Pos.Y;
+      if(!i) {
+	minx = maxx = P1.Pos.X;
+	miny = maxy = P1.Pos.Y;
       }
       minx = DMIN(DMIN(DMIN(minx, P1.Pos.X), P2.Pos.X), P3.Pos.X);
       miny = DMIN(DMIN(DMIN(miny, P1.Pos.Y), P2.Pos.Y), P3.Pos.Y);
@@ -342,9 +427,9 @@ void Draw_Plane_Surface(Surface * s)
     V[3].Pos.X = minx;
     V[3].Pos.Y = maxy;
 
-    for(i = 0; i < 4; i++) {
+    for(int i = 0; i < 4; i++) {
       V[i].Pos.Z = 0.0;
-      put_Z(&V[i], s);
+      putZ(&V[i], s);
     }
 
     n[0] = s->plan[2][0];
@@ -352,73 +437,64 @@ void Draw_Plane_Surface(Surface * s)
     n[2] = s->plan[2][2];
     norme(n);
 
-    k = 0;
-    for(i = 0; i < numPoints; i++) {
-      t = (double)i / (double)(numPoints-1);
-      vv.Pos.X = t * 0.5 * (V[0].Pos.X + V[1].Pos.X) + (1. - t) *
-        0.5 * (V[2].Pos.X + V[3].Pos.X);
-      vv.Pos.Y = t * 0.5 * (V[0].Pos.Y + V[1].Pos.Y) + (1. - t) *
-        0.5 * (V[2].Pos.Y + V[3].Pos.Y);
-      vv.Pos.Z = t * 0.5 * (V[0].Pos.Z + V[1].Pos.Z) + (1. - t) *
-        0.5 * (V[2].Pos.Z + V[3].Pos.Z);
-
-      if(isPointOnPlanarSurface(s, vv.Pos.X, vv.Pos.Y, vv.Pos.Z, n)) {
-        if(!k) {
-          List_Add(s->Orientations, &vv);
-          k = 1;
-        }
-      }
-      else {
-        if(k) {
-          List_Add(s->Orientations, &vv);
-          k = 0;
-        }
+    // compute which parts of the "middle cross" lie inside the surface
+    const int numPoints = 200;
+    for(int cross = 0; cross < 2; cross++) {
+      int k = 0;
+      for(int i = 0; i < numPoints; i++) {
+	double t = (double)i / (double)(numPoints-1);
+	if(!cross){
+	  vv.Pos.X = t * 0.5 * (V[0].Pos.X + V[1].Pos.X) + 
+	    (1. - t) * 0.5 * (V[2].Pos.X + V[3].Pos.X);
+	  vv.Pos.Y = t * 0.5 * (V[0].Pos.Y + V[1].Pos.Y) + 
+	    (1. - t) * 0.5 * (V[2].Pos.Y + V[3].Pos.Y);
+	  vv.Pos.Z = t * 0.5 * (V[0].Pos.Z + V[1].Pos.Z) + 
+	    (1. - t) * 0.5 * (V[2].Pos.Z + V[3].Pos.Z);
+	}
+	else{
+	  vv.Pos.X = t * .5 * (V[0].Pos.X + V[3].Pos.X) + 
+	    (1. - t) * .5 * (V[2].Pos.X + V[1].Pos.X);
+	  vv.Pos.Y = t * .5 * (V[0].Pos.Y + V[3].Pos.Y) + 
+	    (1. - t) * .5 * (V[2].Pos.Y + V[1].Pos.Y);
+	  vv.Pos.Z = t * .5 * (V[0].Pos.Z + V[3].Pos.Z) + 
+	    (1. - t) * .5 * (V[2].Pos.Z + V[1].Pos.Z);
+	}
+	if(isPointOnPlanarSurface(s, vv.Pos.X, vv.Pos.Y, vv.Pos.Z, n)) {
+	  if(!k) {
+	    List_Add(s->Orientations, &vv);
+	    k = 1;
+	  }
+	}
+	else {
+	  if(k) {
+	    List_Add(s->Orientations, &vv);
+	    k = 0;
+	  }
+	}
       }
+      if(k)
+	List_Add(s->Orientations, &vv);
     }
-    if(k)
-      List_Add(s->Orientations, &vv);
 
-    k = 0;
-    for(i = 0; i < numPoints; i++) {
-      t = (double)i / (double)(numPoints-1);
-      vv.Pos.X = t * .5 * (V[0].Pos.X + V[3].Pos.X) + 
-	(1. - t) * .5 * (V[2].Pos.X + V[1].Pos.X);
-      vv.Pos.Y = t * .5 * (V[0].Pos.Y + V[3].Pos.Y) + 
-	(1. - t) * .5 * (V[2].Pos.Y + V[1].Pos.Y);
-      vv.Pos.Z = t * .5 * (V[0].Pos.Z + V[3].Pos.Z) + 
-	(1. - t) * .5 * (V[2].Pos.Z + V[1].Pos.Z);
-      if(isPointOnPlanarSurface(s, vv.Pos.X, vv.Pos.Y, vv.Pos.Z, n)) {
-        if(!k) {
-          List_Add(s->Orientations, &vv);
-          k = 1;
-        }
-      }
-      else {
-        if(k) {
-          List_Add(s->Orientations, &vv);
-          k = 0;
-        }
-      }
+    Msg(STATUS2, "Plane Surface %d (%d cross points)", 
+	s->Num, List_Nbr(s->Orientations));
+    
+    if(!List_Nbr(s->Orientations)){ // add dummy
+      List_Add(s->Orientations, &vv);
+      //printf("surf %d: min=%g %g max=%g %g\n", s->Num, minx, miny, maxx, maxy);
     }
-    if(k)
-      List_Add(s->Orientations, &vv);
-
-    Msg(STATUS2, "Plane Surface %d (%d points)", s->Num, List_Nbr(s->Orientations));
-
-    if(!List_Nbr(s->Orientations)) // add dummy
-      List_Add(s->Orientations, &vv);
 
     CTX.threads_lock = 0;
   }
-
+  
   if(List_Nbr(s->Orientations) > 1) {
 
     if(CTX.geom.surfaces) {
       glEnable(GL_LINE_STIPPLE);
-      glLineStipple(1, 0x0F0F);
+      glLineStipple(1, 0x1F1F);
       gl2psEnable(GL2PS_LINE_STIPPLE);
       glBegin(GL_LINES);
-      for(i = 0; i < List_Nbr(s->Orientations); i++) {
+      for(int i = 0; i < List_Nbr(s->Orientations); i++) {
 	List_Read(s->Orientations, i, &vv);
 	glVertex3d(vv.Pos.X, vv.Pos.Y, vv.Pos.Z);
       }
@@ -430,30 +506,29 @@ void Draw_Plane_Surface(Surface * s)
     if(CTX.geom.surfaces_num) {
       List_Read(s->Orientations, 0, &vv1);
       List_Read(s->Orientations, 1, &vv2);
+      char Num[100];
       sprintf(Num, "%d", s->Num);
-      double offset = 0.5 * CTX.gl_fontsize * CTX.pixel_equiv_x;
+      double offset = 0.3 * CTX.gl_fontsize * CTX.pixel_equiv_x;
       glRasterPos3d((vv2.Pos.X + vv1.Pos.X) / 2. + offset / CTX.s[0],
-                    (vv2.Pos.Y + vv1.Pos.Y) / 2. + offset / CTX.s[1],
-                    (vv2.Pos.Z + vv1.Pos.Z) / 2. + offset / CTX.s[2]);
+		    (vv2.Pos.Y + vv1.Pos.Y) / 2. + offset / CTX.s[1],
+		    (vv2.Pos.Z + vv1.Pos.Z) / 2. + offset / CTX.s[2]);
       Draw_String(Num);
     }
 
     if(CTX.geom.normals) {
       List_Read(s->Orientations, 0, &vv1);
       List_Read(s->Orientations, 1, &vv2);
-      n[0] = s->plan[2][0];
-      n[1] = s->plan[2][1];
-      n[2] = s->plan[2][2];
-      norme(n);
+      double x = (vv1.Pos.X + vv2.Pos.X) / 2.;
+      double y = (vv1.Pos.Y + vv2.Pos.Y) / 2.;
+      double z = (vv1.Pos.Z + vv2.Pos.Z) / 2.;
+      getPlaneSurfaceNormal(s, x, y, z, n);
       n[0] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[0];
       n[1] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[1];
       n[2] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[2];
       glColor4ubv((GLubyte *) & CTX.color.geom.normals);
       Draw_Vector(CTX.vector_type, 0, CTX.arrow_rel_head_radius, 
 		  CTX.arrow_rel_stem_length, CTX.arrow_rel_stem_radius, 
-		  (vv2.Pos.X + vv1.Pos.X) / 2., (vv2.Pos.Y + vv1.Pos.Y) / 2., 
-		  (vv2.Pos.Z + vv1.Pos.Z) / 2., n[0], n[1], n[2], NULL, 
-		  CTX.geom.light);
+		  x, y, z, n[0], n[1], n[2], CTX.geom.light);
     }
 
   }
@@ -464,7 +539,7 @@ void Draw_NonPlane_Surface(Surface * s)
   if(CTX.geom.surfaces) {
     if(s->Typ == MSH_SURF_NURBS) {
       if(CTX.geom.light) glEnable(GL_LIGHTING);
-      glEnable(GL_POLYGON_OFFSET_FILL);
+      if(CTX.polygon_offset) glEnable(GL_POLYGON_OFFSET_FILL);
       GLUnurbsObj *nurb;
       nurb = gluNewNurbsRenderer();
 #if defined(GLU_VERSION_1_3)
@@ -482,7 +557,7 @@ void Draw_NonPlane_Surface(Surface * s)
     }
     else{
       glEnable(GL_LINE_STIPPLE);
-      glLineStipple(1, 0x0F0F);
+      glLineStipple(1, 0x1F1F);
       gl2psEnable(GL2PS_LINE_STIPPLE);
       int N = 50;
       glBegin(GL_LINE_STRIP);
@@ -508,33 +583,29 @@ void Draw_NonPlane_Surface(Surface * s)
     Vertex v = InterpolateSurface(s, 0.5, 0.5, 0, 0);
     char Num[100];
     sprintf(Num, "%d", s->Num);
-    double offset = 0.5 * CTX.gl_fontsize * CTX.pixel_equiv_x;
+    double offset = 0.3 * CTX.gl_fontsize * CTX.pixel_equiv_x;
     glRasterPos3d(v.Pos.X + offset / CTX.s[0],
-                  v.Pos.Y + offset / CTX.s[1],
-                  v.Pos.Z + offset / CTX.s[2]);
+		  v.Pos.Y + offset / CTX.s[1],
+		  v.Pos.Z + offset / CTX.s[2]);
     Draw_String(Num);
   }
 
   if(CTX.geom.normals) {
-    Vertex n1 = InterpolateSurface(s, 0.5, 0.5, 0, 0);
-    Vertex n2 = InterpolateSurface(s, 0.6, 0.5, 0, 0);
-    Vertex n3 = InterpolateSurface(s, 0.5, 0.6, 0, 0);
-    double nx[3], ny[3], n[3];
-    nx[0] = n2.Pos.X - n1.Pos.X;
-    nx[1] = n2.Pos.Y - n1.Pos.Y;
-    nx[2] = n2.Pos.Z - n1.Pos.Z;
-    ny[0] = n3.Pos.X - n1.Pos.X;
-    ny[1] = n3.Pos.Y - n1.Pos.Y;
-    ny[2] = n3.Pos.Z - n1.Pos.Z;
-    prodve(nx, ny, n);
-    norme(n);
+    const double eps = 1.e-3;
+    Vertex v1 = InterpolateSurface(s, 0.5, 0.5, 0, 0);
+    Vertex v2 = InterpolateSurface(s, 0.5 + eps, 0.5, 0, 0);
+    Vertex v3 = InterpolateSurface(s, 0.5, 0.5 + eps, 0, 0);
+    double n[3];
+    normal3points(v1.Pos.X, v1.Pos.Y, v1.Pos.Z, 
+		  v2.Pos.X, v2.Pos.Y, v2.Pos.Z, 
+		  v3.Pos.X, v3.Pos.Y, v3.Pos.Z, n);
     n[0] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[0];
     n[1] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[1];
     n[2] *= CTX.geom.normals * CTX.pixel_equiv_x / CTX.s[2];
     glColor4ubv((GLubyte *) & CTX.color.geom.normals);
     Draw_Vector(CTX.vector_type, 0, CTX.arrow_rel_head_radius, 
 		CTX.arrow_rel_stem_length, CTX.arrow_rel_stem_radius,
-		n1.Pos.X, n1.Pos.Y, n1.Pos.Z, n[0], n[1], n[2], NULL,
+		v1.Pos.X, v1.Pos.Y, v1.Pos.Z, n[0], n[1], n[2],
 		CTX.geom.light);
   }
 }
@@ -543,39 +614,42 @@ void Draw_Surface(void *a, void *b)
 {
   Surface *s = *(Surface **) a;
 
-  if(!s || !s->Support || !(s->Visible & VIS_GEOM) || s->Dirty)
+  if(!s || !s->Support || !(s->Visible & VIS_GEOM))
     return;
 
   if(CTX.render_mode == GMSH_SELECT) {
-    glLoadName(2);
+    glPushName(2);
     glPushName(s->Num);
   }
 
-  if(s->ipar[4]) {
-    glLineWidth(CTX.geom.line_sel_width);
-    gl2psLineWidth(CTX.geom.line_sel_width *
+  if(s->ipar[4] > 0) {
+    glLineWidth(CTX.geom.line_sel_width / 2.);
+    gl2psLineWidth(CTX.geom.line_sel_width / 2. *
 		   CTX.print.eps_line_width_factor);
     glColor4ubv((GLubyte *) & CTX.color.geom.surface_sel);
   }
   else {
-    glLineWidth(CTX.geom.line_width);
-    gl2psLineWidth(CTX.geom.line_width * CTX.print.eps_line_width_factor);
+    glLineWidth(CTX.geom.line_width / 2.);
+    gl2psLineWidth(CTX.geom.line_width / 2. * CTX.print.eps_line_width_factor);
     glColor4ubv((GLubyte *) & CTX.color.geom.surface);
   }
 
-  if(s->Typ == MSH_SURF_STL) {
-    Tree_Action(s->STL->Simplexes, Draw_Mesh_Triangle);
+  if(THEM->bds){
+    Draw_Polygonal_Surface(s);
   }
-  else if(s->Typ == MSH_SURF_DISCRETE) {
-    Tree_Action(s->Simplexes, Draw_Mesh_Triangle);
+  else if(s->Typ == MSH_SURF_DISCRETE){
+    // do nothing: we draw the elements in the mesh drawing routines
   }
-  else if(s->Typ == MSH_SURF_PLAN)
+  else if(s->Typ == MSH_SURF_PLAN){
     Draw_Plane_Surface(s);
-  else
+  }
+  else{
     Draw_NonPlane_Surface(s);
+  }
 
   if(CTX.render_mode == GMSH_SELECT) {
     glPopName();
+    glPopName();
   }
 }
 
@@ -609,16 +683,19 @@ void HighlightEntity(Vertex * v, Curve * c, Surface * s, int permanent)
   Curve *cc;
   char Message[256], temp[256];
 
-  if(permanent){
-    // we want to draw incrementally (in-between to "Draw()" calls):
-    // we need to make sure that the opengl context is set correctly
-    SetOpenglContext();
-  }
+  // Note: in Gmsh < 1.61, we used to draw permanent highlights (the
+  // "red" selected lines, etc.) using incremental drawing, i.e., by
+  // drawing "over" the current picture in-between Draw() calls. This
+  // was fine for simple overlays on points and lines (that we could
+  // draw with a slightly larger width so that they would cover the
+  // old ones), but it does not work well when drawing surfaces,
+  // post-pro views, etc. And since real cross-platform overlays are
+  // unmanageable, the best solution is actually to redraw the whole
+  // scene. This is the approach we follow now.
 
   if(v) {
     if(permanent){
       v->Frozen = 1;
-      Draw_Geo_Point(&v,NULL);
     }
     else{
       Msg(STATUS1N, "Point %d {%.5g,%.5g,%.5g} (%.5g)", v->Num, v->Pos.X,
@@ -627,42 +704,56 @@ void HighlightEntity(Vertex * v, Curve * c, Surface * s, int permanent)
   }
   else if(c) {
     if(permanent){
-      c->ipar[3] = 1;
-      Draw_Curve(&c,NULL);
+      c->ipar[3] = 2;
     }
     else{
-      Msg(STATUS1N, "Curve %d  {%d->%d}", c->Num, c->beg->Num, c->end->Num);
+      if(c->beg && c->end)
+	Msg(STATUS1N, "Curve %d {%d->%d}", c->Num, c->beg->Num, c->end->Num);
+      else
+	Msg(STATUS1N, "Curve %d", c->Num);
     }
   }
   else if(s) {
     if(permanent){
-      s->ipar[4] = 1;
-      Draw_Surface(&s,NULL);
+      s->ipar[4] = 2;
     }
     else{
-      sprintf(Message, "Surface %d {", s->Num);
       int nbg = List_Nbr(s->Generatrices);
-      if(nbg < 10) {
-	for(int i = 0; i < nbg; i++) {
-	  List_Read(s->Generatrices, i, &cc);
-	  if(!i)
-	    sprintf(temp, "%d", cc->Num);
-	  else
-	    sprintf(temp, ",%d", cc->Num);
-	  strcat(Message, temp);
+      if(nbg){
+	sprintf(Message, "Surface %d {", s->Num);
+	if(nbg < 10) {
+	  for(int i = 0; i < nbg; i++) {
+	    List_Read(s->Generatrices, i, &cc);
+	    if(!i)
+	      sprintf(temp, "%d", cc->Num);
+	    else
+	      sprintf(temp, ",%d", cc->Num);
+	    strcat(Message, temp);
+	  }
 	}
+	else {
+	  strcat(Message, "...");
+	}
+	strcat(Message, "}");
+	Msg(STATUS1N, Message);
       }
-      else {
-	strcat(Message, "...");
+      else if (THEM->bds){
+	BDS_GeomEntity *g = THEM->bds->get_geom(s->Num, 2);
+	if (g && g->surf)
+	  Msg(STATUS1N, "Surface %d (%s)", s->Num,g->surf->nameOf().c_str());
+	else
+	  Msg(STATUS1N, "Surface %d (unknown type)", s->Num);
+      }
+      else{
+	Msg(STATUS1N, "Surface %d", s->Num);
       }
-      strcat(Message, "}");
-      Msg(STATUS1N, Message);
     }
   }
   else{
     if(!permanent)
       Msg(STATUS1N, " ");
   }
+
 }
 
 void HighlightEntityNum(int v, int c, int s, int permanent)
@@ -686,23 +777,20 @@ void HighlightEntityNum(int v, int c, int s, int permanent)
 
 void ZeroHighlightPoint(void *a, void *b)
 {
-  Vertex *v;
-  v = *(Vertex **) a;
+  Vertex *v = *(Vertex **) a;
   v->Frozen = 0;
 }
 
 void ZeroHighlightCurve(void *a, void *b)
 {
-  Curve *c;
-  c = *(Curve **) a;
-  c->ipar[3] = 0;
+  Curve *c = *(Curve **) a;
+  c->ipar[3] = -2;
 }
 
 void ZeroHighlightSurface(void *a, void *b)
 {
-  Surface *s;
-  s = *(Surface **) a;
-  s->ipar[4] = 0;
+  Surface *s = *(Surface **) a;
+  s->ipar[4] = -2;
 }
 
 void ZeroHighlight(Mesh * m)
@@ -712,6 +800,13 @@ void ZeroHighlight(Mesh * m)
   Tree_Action(m->Surfaces, ZeroHighlightSurface);
 }
 
+void ZeroHighlightEntity(Vertex * v, Curve * c, Surface * s)
+{
+  if(v) ZeroHighlightPoint(&v, NULL);
+  if(c) ZeroHighlightCurve(&c, NULL);
+  if(s) ZeroHighlightSurface(&s, NULL);
+}
+
 void ZeroHighlightEntityNum(int v, int c, int s)
 {
   if(v) {
