@@ -1,6 +1,6 @@
-// $Id: 2D_Mesh.cpp,v 1.58 2004/05/26 00:33:37 geuzaine Exp $
+// $Id: 2D_Mesh.cpp,v 1.81.2.1 2006/04/16 02:42:42 geuzaine Exp $
 //
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 #include "Create.h"
 #include "2D_Mesh.h"
 #include "Context.h"
+#include "Interpolation.h"
 
 extern Mesh *THEM;
 extern Context_T CTX;
@@ -36,7 +37,6 @@ PointRecord *gPointArray;
 DocRecord *BGMESH, *FGMESH;
 double LC2D;
 
-static int is_3D = 0;
 static Surface *THESURFACE, *THESUPPORT;
 
 void ProjetteSurface(void *a, void *b)
@@ -65,12 +65,10 @@ void Plan_Moyen(void *data, void *dum)
   int i, j;
   Vertex *v;
   Curve *pC;
-  Surface **pS, *s;
   static List_T *points;
   static int deb = 1;
 
-  pS = (Surface **) data;
-  s = *pS;
+  Surface *s = *(Surface **) data;
 
   if(deb) {
     points = List_Create(10, 10, sizeof(Vertex *));
@@ -111,6 +109,10 @@ int Calcule_Contours(Surface * s)
   l = List_Create(10, 10, sizeof(Vertex *));
   for(i = 0; i < List_Nbr(s->Generatrices); i++) {
     List_Read(s->Generatrices, i, &c);
+    if(!List_Nbr(c->Vertices)) {
+      Msg(GERROR, "Curve %d has no mesh", c->Num);
+      return 0;
+    }
     if(!List_Nbr(l)) {
       List_Read(c->Vertices, 0, &first);
     }
@@ -274,7 +276,7 @@ int mesh_domain(ContourPeek * ListContours, int numcontours,
 
   if(!numcontours) {
     Msg(GERROR, "No contour");
-    return 0;
+    return -1;
   }
 
   numact = 0;
@@ -302,7 +304,7 @@ int mesh_domain(ContourPeek * ListContours, int numcontours,
   Conversion(doc);
   remove_all_dlist(doc->numPoints, doc->points);
 
-  if(!is_3D || CTX.mesh.constrained_bgmesh)
+  if(THEM->BackgroundMeshType != ONFILE || CTX.mesh.constrained_bgmesh)
     BGMESH = doc;
   else
     BGMESH = NULL;
@@ -324,7 +326,6 @@ int mesh_domain(ContourPeek * ListContours, int numcontours,
       List_Add(del_L, &del_P);
   }
   doc->numTriangles = List_Nbr(del_L);
-
   verify_edges(del_L, ListContours, numcontours);
   verify_inside(doc->delaunay, doc->numTriangles);
 
@@ -455,7 +456,7 @@ int mesh_domain(ContourPeek * ListContours, int numcontours,
     }
 
     *numpoints = doc->numPoints;
-    Insert_Point(pt, numpoints, &numaloc, doc, BGMESH, is_3D);
+    Insert_Point(pt, numpoints, &numaloc, doc, BGMESH);
     doc->points = gPointArray;
     doc->numPoints = *numpoints;
 
@@ -565,16 +566,7 @@ void Maillage_Automatique_VieuxCode(Surface * pS, Mesh * m, int ori)
   Simplex *s;
   double Xmin = 0., Xmax = 0., Ymin = 0., Ymax = 0.;
 
-
-  if(m->BGM.Typ == WITHPOINTS) {
-    is_3D = 0;
-  }
-  else {
-    is_3D = 1;
-  }
-
-  liste =
-    (ContourPeek *) Malloc(List_Nbr(pS->Contours) * sizeof(ContourPeek));
+  liste = (ContourPeek *) Malloc(List_Nbr(pS->Contours) * sizeof(ContourPeek));
 
   k = 0;
 
@@ -624,9 +616,10 @@ void Maillage_Automatique_VieuxCode(Surface * pS, Mesh * m, int ori)
     liste[i] = cp;
   }
 
+  int res_mesh_domain = 1;
   if(pS->Method)
-    mesh_domain(liste, List_Nbr(pS->Contours), &M, &N,
-                (CTX.mesh.initial_only == 2));
+    res_mesh_domain = mesh_domain(liste, List_Nbr(pS->Contours), &M, &N,
+				  (CTX.mesh.initial_only == 2));
 
   for(i = 0; i < M.numpoints; i++) {
     if(gPointArray[i].initial < 0) {
@@ -681,7 +674,9 @@ void Maillage_Automatique_VieuxCode(Surface * pS, Mesh * m, int ori)
     //xxxxFree(M.listdel[i]);
 
   }
-  Free(M.listdel);
+  
+  if(res_mesh_domain >= 0)
+    Free(M.listdel);
   Free(gPointArray);
 }
 
@@ -714,33 +709,19 @@ void filldel(Delaunay * deladd, int aa, int bb, int cc,
   deladd->v.voisin2 = NULL;
   deladd->v.voisin3 = NULL;
 
-  CircumCircle(points[aa].where.h,
-               points[aa].where.v,
-               points[bb].where.h,
-               points[bb].where.v,
-               points[cc].where.h,
-               points[cc].where.v, &deladd->t.xc, &deladd->t.yc);
+  CircumCircle(points[aa].where.h, points[aa].where.v,
+               points[bb].where.h, points[bb].where.v,
+               points[cc].where.h, points[cc].where.v, 
+	       &deladd->t.xc, &deladd->t.yc);
 
   pt2.h = deladd->t.xc;
   pt2.v = deladd->t.yc;
-  if(!is_3D) {
-    if(mesh) {
-      newqual = find_quality(pt2, mesh);
-    }
-    else {
-      newqual =
-        (points[aa].quality + points[bb].quality + points[cc].quality) / 3.;
-    }
+
+  if(THEM->BackgroundMeshType == ONFILE) {
     v = Create_Vertex(-1, pt2.h, pt2.v, 0.0, 0.0, 0.0);
     Calcule_Z_Plan(&v, &dum);
     Projette_Inverse(&v, &dum);
-    Free_Vertex(&v, 0);
-  }
-  else {
-    v = Create_Vertex(-1, pt2.h, pt2.v, 0.0, 0.0, 0.0);
-    Calcule_Z_Plan(&v, &dum);
-    Projette_Inverse(&v, &dum);
-    qual = Lc_XYZ(v->Pos.X, v->Pos.Y, v->Pos.Z, THEM);
+    qual = BGMXYZ(v->Pos.X, v->Pos.Y, v->Pos.Z);
     if(CTX.mesh.constrained_bgmesh) {
       if(mesh) {
         newqual = MIN(qual, find_quality(pt2, mesh));
@@ -756,6 +737,17 @@ void filldel(Delaunay * deladd, int aa, int bb, int cc,
       newqual = qual;
     Free_Vertex(&v, 0);
   }
+  else{
+    if(mesh)
+      newqual = find_quality(pt2, mesh);
+    else
+      newqual = (points[aa].quality + points[bb].quality + 
+		 points[cc].quality) / 3.;
+    v = Create_Vertex(-1, pt2.h, pt2.v, 0.0, 0.0, 0.0);
+    Calcule_Z_Plan(&v, &dum);
+    Projette_Inverse(&v, &dum);
+    Free_Vertex(&v, 0);
+  }
 
   deladd->t.quality_value =
     sqrt((deladd->t.xc - points[cc].where.h) * (deladd->t.xc -
@@ -772,27 +764,97 @@ void ActionEndTheCurve(void *a, void *b)
   End_Curve(c);
 }
 
+void ActionInvertTri(void *a, void *b)
+{
+  Simplex *s = *(Simplex **) a;
+  Vertex *tmp = s->V[1];
+  s->V[1] = s->V[2];
+  s->V[2] = tmp;
+}
+
+void ActionInvertQua(void *a, void *b)
+{
+  Quadrangle *q = *(Quadrangle **) a;
+  Vertex *tmp = q->V[1];
+  q->V[1] = q->V[3];
+  q->V[3] = tmp;
+}
+
+// A little routine that re-orients the surface meshes so that they
+// match the geometrical orientations. It would be better to generate
+// the meshes correctly in the first place (!), but it's actually
+// pretty nice to do it here since we use many algorithms (some
+// closed) and since the mean plane computation doesn't take the
+// original orientation into account.
+
+void ReOrientSurfaceMesh(Surface *s)
+{
+  if(!List_Nbr(s->Generatrices))
+    return;
+
+  Curve *c;
+  List_Read(s->Generatrices, 0, &c);
+
+  if(List_Nbr(c->Vertices) < 2)
+    return;
+
+  // take the first edge (even if there are holes, we know that this
+  // edge belongs to the exterior loop, due to the ordering of
+  // s->Generatrices)
+  Vertex *v1, *v2;
+  List_Read(c->Vertices, 0, &v1);
+  List_Read(c->Vertices, 1, &v2);
+  Edge edge;
+  edge.V[0] = v1;
+  edge.V[1] = v2;
+
+  if(Tree_Nbr(s->Simplexes)){
+    EdgesContainer edges;
+    edges.AddSimplexTree(s->Simplexes);
+    Edge *e = (Edge*)Tree_PQuery(edges.AllEdges, &edge);
+    if(e && e->V[0]->Num == v2->Num && e->V[1]->Num == v1->Num){
+      Tree_Action(s->Simplexes, ActionInvertTri);
+      if(Tree_Nbr(s->Quadrangles)) // for partially recombined meshes
+	Tree_Action(s->Quadrangles, ActionInvertQua);
+    }
+  }
+
+  if(Tree_Nbr(s->Quadrangles)){
+    EdgesContainer edges;
+    edges.AddQuadrangleTree(s->Quadrangles);
+    Edge *e = (Edge*)Tree_PQuery(edges.AllEdges, &edge);
+    if(e && e->V[0]->Num == v2->Num && e->V[1]->Num == v1->Num){
+      Tree_Action(s->Quadrangles, ActionInvertQua);
+      if(Tree_Nbr(s->Simplexes)) // for partially recombined meshes
+	Tree_Action(s->Simplexes, ActionInvertTri);
+    }
+  }
+}
+
 void Maillage_Surface(void *data, void *dum)
 {
-  Surface **pS, *s;
   Tree_T *tnxe;
   int ori;
 
-  pS = (Surface **) data;
-  s = *pS;
+  Surface *s = *(Surface **) data;
 
   if(!s->Support)
     return;
 
-  if(s->Dirty) {
-    Msg(INFO, "Not meshing dirty Surface %d", s->Num);
-    return;
-  }
+  Msg(STATUS3, "Meshing surface %d", s->Num);
 
-  if(s->Typ == MSH_SURF_STL) {
-    Msg(GERROR, "Remeshing of STL surfaces is not implemented (yet)");
+  int tag = MeshDiscreteSurface(s);
+
+  if(tag == 1){
+    Tree_Action(THEM->Points, PutVertex_OnSurf);
+    Tree_Action(s->Vertices, PutVertex_OnSurf);
+    Tree_Action(s->Vertices, Add_In_Mesh);
     return;
   }
+  else if (tag == 2)
+    {
+      return;
+    }
 
   THESUPPORT = s->Support;
   THESURFACE = s;
@@ -804,20 +866,24 @@ void Maillage_Surface(void *data, void *dum)
     Tree_Delete(s->Vertices);
   s->Vertices = Tree_Create(sizeof(Vertex *), compareVertex);
 
-  Msg(STATUS3, "Meshing surface %d", s->Num);
-
   if(MeshTransfiniteSurface(s) ||
      MeshEllipticSurface(s) ||
      MeshCylindricalSurface(s) ||
-     MeshParametricSurface(s) || 
+     MeshParametricSurface(s) ||
      Extrude_Mesh(s)) {
     Tree_Action(THEM->Points, PutVertex_OnSurf);
     Tree_Action(s->Vertices, PutVertex_OnSurf);
     Tree_Action(s->Vertices, Add_In_Mesh);
   }
   else{
+
+    if(!List_Nbr(s->Generatrices)){
+      Msg(GERROR, "Cannot mesh surface with no boundary");
+      return;
+    }
+
     int TypSurface = s->Typ;
-    Plan_Moyen(pS, dum);
+    Plan_Moyen(&s, dum);
     Tree_Action(THEM->Points, Freeze_Vertex);
     Tree_Action(s->Vertices, Freeze_Vertex);
     Tree_Action(THEM->Points, Projette_Plan_Moyen);
@@ -828,12 +894,12 @@ void Maillage_Surface(void *data, void *dum)
 
     ori = Calcule_Contours(s);
     
-    if(CTX.mesh.algo == DELAUNAY_ISO)
+    if(CTX.mesh.algo2d == DELAUNAY_ISO)
       Maillage_Automatique_VieuxCode(s, THEM, ori);
-    else if(CTX.mesh.algo == DELAUNAY_ANISO)
+    else if(CTX.mesh.algo2d == DELAUNAY_ANISO)
       AlgorithmeMaillage2DAnisotropeModeJF(s);
     else
-      Mesh_Shewchuk(s);
+      Mesh_Triangle(s);
     
     if(CTX.mesh.nb_smoothing) {
       Msg(STATUS3, "Smoothing surface %d", s->Num);
@@ -874,6 +940,8 @@ void Maillage_Surface(void *data, void *dum)
     Tree_Action(THEM->Curves, ActionEndTheCurve);
     End_Surface(s->Support, 0);
     End_Surface(s, 0);
+    
+    ReOrientSurfaceMesh(s);
   }
 
 }
