@@ -1,6 +1,6 @@
-// $Id: 3D_Extrude_Old.cpp,v 1.27 2004/05/25 04:10:04 geuzaine Exp $
+// $Id: 3D_Extrude_Old.cpp,v 1.41 2006/01/29 22:53:41 geuzaine Exp $
 //
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -19,17 +19,17 @@
 // 
 // Please report all bugs and problems to <gmsh@geuz.org>.
 
-// This is the old extrusion mesh generator -> only available through
-// the command line option -extrude (w/o -recombine). This mesh
+// This is the old extrusion mesh generator, only available through
+// the command line option -extrude (w/ or w/o -recombine). This mesh
 // generator pre-supposes a definition of surfaces in the XY plane,
 // and will extrude everything along the Z axis, taking parameters
-// interactively from standard input, e.g.
+// interactively from standard input, e.g.,
 //
 // gmsh test -extrude -recombine < params.ext
 //
 // The progression ratio defines a geometric progression for the
 // definition of the elements heights: a factor of 2 means that the
-// next element will be twice as high than the preceding one.
+// next element will be twice as high as the previous one.
 //
 // All geometrical entities are automatically numbered:
 //
@@ -53,32 +53,26 @@
 #include "Create.h"
 
 extern Context_T CTX;
-extern Mesh *LOCAL, *THEM;
+extern Mesh *THEM;
 
-static Tree_T *Tree_Ares, *Tree_Swaps;
+static Tree_T *Tree_Ares = NULL, *Tree_Swaps = NULL;
+static Volume *THEV;
+static int BAD_TETS, TRY_INVERSE, NbLayer;
+static int NbElmLayer[MAXLAYERS];
+static int ZonLayer[MAXLAYERS];
+static int LineLayer[MAXLAYERS + 1];
+static int SurfLayer[MAXLAYERS + 1];
+static double hLayer[MAXLAYERS];
+static double parLayer[MAXLAYERS];
 
-Surface *THES;
-Volume *THEV;
-int TEST_IS_ALL_OK, NbLayer;
-int NbElmLayer[MAXLAYERS];
-int ZonLayer[MAXLAYERS];
-int LineLayer[MAXLAYERS + 1];
-int SurfLayer[MAXLAYERS + 1];
-double hLayer[MAXLAYERS];
-double parLayer[MAXLAYERS];
-
-
-typedef struct
-{
+typedef struct{
   int a, b;
-}
-nxn;
+} nxn;
 
 static int compnxn(const void *a, const void *b)
 {
-  nxn *q, *w;
-  q = (nxn *) a;
-  w = (nxn *) b;
+  nxn *q = (nxn *) a;
+  nxn *w = (nxn *) b;
   if(q->a > w->a)
     return 1;
   if(q->a < w->a)
@@ -92,33 +86,50 @@ static int compnxn(const void *a, const void *b)
 
 static void InitExtrudeParams(void)
 {
-  FILE *file;
-  int i;
+  char *ret;
+  char str[256];
 
-  printf("Number of layers: ");
-  scanf("%d", &NbLayer);
+  printf("Number of layers (default=1): ");
+  ret = fgets(str, sizeof(str), stdin);
+  if(!ret || !strlen(str) || !strcmp(str, "\n"))
+    NbLayer = 1;
+  else
+    NbLayer = atoi(str);
+  
   if(NbLayer > MAXLAYERS){
     Msg(GERROR, "Max number of layer (%d) exceeded", MAXLAYERS);
     NbLayer = MAXLAYERS;
   }
 
-  file = fopen("xtrude", "w");
+  FILE *file = fopen(".gmsh-extrude", "w");
 
   if(file)
     fprintf(file, "%d\n", NbLayer);
-  for(i = 0; i < NbLayer; i++) {
-    printf("Number of elements in layer %d: ", i + 1);
-    scanf("%d", &NbElmLayer[i]);
+  for(int i = 0; i < NbLayer; i++) {
+    printf("Number of elements in layer %d (default=1): ", i + 1);
+    ret = fgets(str, sizeof(str), stdin);
+    if(!ret || !strlen(str) || !strcmp(str, "\n"))
+      NbElmLayer[i] = 1;
+    else
+      NbElmLayer[i] = atoi(str);
     if(file)
       fprintf(file, "%d\n", NbElmLayer[i]);
 
-    printf("Depth of layer %d: ", i + 1);
-    scanf("%lf", &hLayer[i]);
+    printf("Depth of layer %d (default=1.0): ", i + 1);
+    ret = fgets(str, sizeof(str), stdin);
+    if(!ret || !strlen(str) || !strcmp(str, "\n"))
+      hLayer[i] = 1.0;
+    else
+      hLayer[i] = atof(str);
     if(file)
       fprintf(file, "%g\n", hLayer[i]);
 
-    printf("Progresion ratio for layer %d: ", i + 1);
-    scanf("%lf", &parLayer[i]);
+    printf("Progression ratio for layer %d (default=1.0): ", i + 1);
+    ret = fgets(str, sizeof(str), stdin);
+    if(!ret || !strlen(str) || !strcmp(str, "\n"))
+      parLayer[i] = 1.0;
+    else
+      parLayer[i] = atof(str);
     if(file)
       fprintf(file, "%g\n", parLayer[i]);
   }
@@ -127,9 +138,6 @@ static void InitExtrudeParams(void)
     fflush(file);
     fclose(file);
   }
-
-  Tree_Ares = Tree_Create(sizeof(nxn), compnxn);
-  Tree_Swaps = Tree_Create(sizeof(nxn), compnxn);
 }
 
 static int are_exist(Vertex * v1, Vertex * v2, Tree_T * t)
@@ -156,17 +164,15 @@ static void are_del(Vertex * v1, Vertex * v2, Tree_T * t)
   Tree_Suppress(t, &n);
 }
 
-
 static void Extrude_Simplex_Phase1(void *data, void *dum)
 {
-  int i, j, k;
   Vertex *v1, *v2, *v3, *v4, *v5, *v6;
 
   Simplex *s = *(Simplex **) data;
 
-  k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
+  int k = 0;
+  for(int i = 0; i < NbLayer; i++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       List_Read(s->V[0]->Extruded_Points, k, &v1);
       List_Read(s->V[1]->Extruded_Points, k, &v2);
       List_Read(s->V[2]->Extruded_Points, k, &v3);
@@ -186,31 +192,30 @@ static void Extrude_Simplex_Phase1(void *data, void *dum)
 
 static void Extrude_Simplex_Phase3(void *data, void *dum)
 {
-  Simplex  *news;
+  Simplex *news;
   Prism *newp;
-  int i, j, k;
   Vertex *v1, *v2, *v3, *v4, *v5, *v6;
 
   Simplex *s = *(Simplex **) data;
 
-  k = 0;
-  for(i = 0; i <= NbLayer; i++) {
+  int k = 0;
+  for(int i = 0; i <= NbLayer; i++) {
     if(SurfLayer[i]) {
       List_Read(s->V[0]->Extruded_Points, k, &v1);
       List_Read(s->V[1]->Extruded_Points, k, &v2);
       List_Read(s->V[2]->Extruded_Points, k, &v3);
-      news = Create_Simplex(v1, v2, v3, NULL);
+      Simplex *news = Create_Simplex(v1, v2, v3, NULL);
       news->iEnt = SurfLayer[i];
       Tree_Add(THEV->Simp_Surf, &news);
     }
-    for(j = 0; j < NbElmLayer[i]; j++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       k++;
     }
   }
 
   k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
+  for(int i = 0; i < NbLayer; i++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       List_Read(s->V[0]->Extruded_Points, k, &v1);
       List_Read(s->V[1]->Extruded_Points, k, &v2);
       List_Read(s->V[2]->Extruded_Points, k, &v3);
@@ -227,7 +232,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
         }
         else {
           if(are_exist(v4, v2, Tree_Ares) &&
-             are_exist(v5, v3, Tree_Ares) && are_exist(v4, v3, Tree_Ares)) {
+             are_exist(v5, v3, Tree_Ares) && 
+	     are_exist(v4, v3, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v4);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -239,7 +245,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
             Tree_Add(THEV->Simplexes, &news);
           }
           if(are_exist(v4, v2, Tree_Ares) &&
-             are_exist(v2, v6, Tree_Ares) && are_exist(v4, v3, Tree_Ares)) {
+             are_exist(v2, v6, Tree_Ares) && 
+	     are_exist(v4, v3, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v4);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -251,7 +258,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
             Tree_Add(THEV->Simplexes, &news);
           }
           if(are_exist(v4, v2, Tree_Ares) &&
-             are_exist(v2, v6, Tree_Ares) && are_exist(v6, v1, Tree_Ares)) {
+             are_exist(v2, v6, Tree_Ares) && 
+	     are_exist(v6, v1, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v6);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -263,7 +271,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
             Tree_Add(THEV->Simplexes, &news);
           }
           if(are_exist(v5, v1, Tree_Ares) &&
-             are_exist(v5, v3, Tree_Ares) && are_exist(v4, v3, Tree_Ares)) {
+             are_exist(v5, v3, Tree_Ares) && 
+	     are_exist(v4, v3, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v5);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -275,7 +284,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
             Tree_Add(THEV->Simplexes, &news);
           }
           if(are_exist(v5, v1, Tree_Ares) &&
-             are_exist(v5, v3, Tree_Ares) && are_exist(v6, v1, Tree_Ares)) {
+             are_exist(v5, v3, Tree_Ares) && 
+	     are_exist(v6, v1, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v5);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -287,7 +297,8 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
             Tree_Add(THEV->Simplexes, &news);
           }
           if(are_exist(v5, v1, Tree_Ares) &&
-             are_exist(v2, v6, Tree_Ares) && are_exist(v6, v1, Tree_Ares)) {
+             are_exist(v2, v6, Tree_Ares) && 
+	     are_exist(v6, v1, Tree_Ares)) {
             news = Create_Simplex(v1, v2, v3, v6);
             news->iEnt = ZonLayer[i];
             Tree_Add(THEV->Simplexes, &news);
@@ -306,8 +317,6 @@ static void Extrude_Simplex_Phase3(void *data, void *dum)
 
 static void Extrude_Quadrangle_Phase3(void *data, void *dum)
 {
-  int i, j, k;
-  Vertex *v1, *v2, *v3, *v4, *v5, *v6, *v7, *v8;
 
   Quadrangle *q = *(Quadrangle **) data;
 
@@ -316,8 +325,10 @@ static void Extrude_Quadrangle_Phase3(void *data, void *dum)
     return;
   }
 
-  k = 0;
-  for(i = 0; i <= NbLayer; i++) {
+  Vertex *v1, *v2, *v3, *v4, *v5, *v6, *v7, *v8;
+
+  int k = 0;
+  for(int i = 0; i <= NbLayer; i++) {
     if(SurfLayer[i]) {
       List_Read(q->V[0]->Extruded_Points, k, &v1);
       List_Read(q->V[1]->Extruded_Points, k, &v2);
@@ -327,14 +338,14 @@ static void Extrude_Quadrangle_Phase3(void *data, void *dum)
       newq->iEnt = SurfLayer[i];
       Tree_Add(THEV->Quad_Surf, &newq);
     }
-    for(j = 0; j < NbElmLayer[i]; j++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       k++;
     }
   }
 
   k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
+  for(int i = 0; i < NbLayer; i++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       List_Read(q->V[0]->Extruded_Points, k, &v1);
       List_Read(q->V[1]->Extruded_Points, k, &v2);
       List_Read(q->V[2]->Extruded_Points, k, &v3);
@@ -356,14 +367,12 @@ static void Extrude_Quadrangle_Phase3(void *data, void *dum)
 
 static void Extrude_Simplex_Phase2(void *data, void *dum)
 {
-  int i, j, k;
-  Vertex *v1, *v2, *v3, *v4, *v5, *v6;
-
   Simplex *s = *(Simplex **) data;
 
-  k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
+  int k = 0;
+  for(int i = 0; i < NbLayer; i++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
+      Vertex *v1, *v2, *v3, *v4, *v5, *v6;
       List_Read(s->V[0]->Extruded_Points, k, &v1);
       List_Read(s->V[1]->Extruded_Points, k, &v2);
       List_Read(s->V[2]->Extruded_Points, k, &v3);
@@ -372,8 +381,9 @@ static void Extrude_Simplex_Phase2(void *data, void *dum)
       List_Read(s->V[2]->Extruded_Points, k + 1, &v6);
       k++;
       if(are_exist(v4, v2, Tree_Ares) &&
-         are_exist(v5, v3, Tree_Ares) && are_exist(v1, v6, Tree_Ares)) {
-        TEST_IS_ALL_OK++;
+         are_exist(v5, v3, Tree_Ares) && 
+	 are_exist(v1, v6, Tree_Ares)) {
+        BAD_TETS++;
         if(!are_exist(v4, v2, Tree_Swaps)) {
           are_del(v4, v2, Tree_Ares);
           are_cree(v1, v5, Tree_Ares);
@@ -394,8 +404,9 @@ static void Extrude_Simplex_Phase2(void *data, void *dum)
         }
       }
       else if(are_exist(v1, v5, Tree_Ares) &&
-              are_exist(v2, v6, Tree_Ares) && are_exist(v4, v3, Tree_Ares)) {
-        TEST_IS_ALL_OK++;
+              are_exist(v2, v6, Tree_Ares) && 
+	      are_exist(v4, v3, Tree_Ares)) {
+        BAD_TETS++;
         if(!are_exist(v1, v5, Tree_Swaps)) {
           are_del(v1, v5, Tree_Ares);
           are_cree(v4, v2, Tree_Ares);
@@ -421,56 +432,52 @@ static void Extrude_Simplex_Phase2(void *data, void *dum)
 
 static void Extrude_Vertex(void *data, void *dum)
 {
-  Vertex **pV, *v, *newv;
-  int i, j;
-  double h, a;
+  Vertex *v = *(Vertex **) data;
 
-  pV = (Vertex **) data;
-  v = *pV;
   if(v->Extruded_Points)
     return;
+
   v->Extruded_Points = List_Create(NbLayer, 1, sizeof(Vertex *));
   List_Add(v->Extruded_Points, &v);
-  h = 0.0;
 
   //printf("-extruding vertex %d %p\n", v->Num, v);
 
-  for(i = 0; i < NbLayer; i++) {
+  double h = 0.0;
+
+  for(int i = 0; i < NbLayer; i++) {
 
     // Geometric progression ar^i
     // Sum of n (=NbElmLayer[i]) terms = hLayer[i] = a (r^n-1)/(r-1)
-
+    
+    double a;
     if(parLayer[i] == 1.)
       a = hLayer[i] / (double)NbElmLayer[i];
     else
-      a =
-        hLayer[i] * (parLayer[i] - 1.) / (pow(parLayer[i], NbElmLayer[i]) -
-                                          1.);
-
-    for(j = 0; j < NbElmLayer[i]; j++) {
-
+      a = hLayer[i] * (parLayer[i] - 1.) / 
+	(pow(parLayer[i], NbElmLayer[i]) - 1.);
+    
+    for(int j = 0; j < NbElmLayer[i]; j++) {
+     
       //h += hLayer[i]/(double)NbElmLayer[i];
 
       h += a * pow(parLayer[i], j);
 
-      newv =
-        Create_Vertex(++THEM->MaxPointNum, v->Pos.X, v->Pos.Y, v->Pos.Z + h,
-                      v->lc, v->u);
+      Vertex *newv = Create_Vertex(++THEM->MaxPointNum, v->Pos.X, v->Pos.Y, 
+			   v->Pos.Z + h, v->lc, v->u);
       Tree_Add(THEM->Vertices, &newv);
       List_Add(v->Extruded_Points, &newv);
     }
   }
 }
 
-
 static void Extrude_Surface1(void *data, void *dum)
 {
   if(!NbLayer)
     return;
   Surface *s = *(Surface **) data;
-  THES = s;
 
   Tree_Action(s->Vertices, Extrude_Vertex);
+
   if(!CTX.mesh.oldxtrude_recombine)
     Tree_Action(s->Simplexes, Extrude_Simplex_Phase1);
 }
@@ -480,28 +487,27 @@ static void Extrude_Surface2(void *data, void *dum)
   if(!NbLayer)
     return;
   Surface *s = *(Surface **) data;
-  THES = s;
 
-  Tree_Action(s->Simplexes, Extrude_Simplex_Phase2);
+  // we could try a whole bunch of orderings...
+  if(TRY_INVERSE)
+    Tree_Action_Inverse(s->Simplexes, Extrude_Simplex_Phase2);
+  else
+    Tree_Action(s->Simplexes, Extrude_Simplex_Phase2);
 }
 
-
 static void Extrude_Surface3(void *data, void *dum)
 {
-  int i;
   if(!NbLayer)
     return;
 
   Surface *s = *(Surface **) data;
-  THES = s;
 
-  /* Numerotation automatique des entites physiques */
   Msg(INFO, "Extruding Surface %d", s->Num);
-  for(i = 0; i < NbLayer; i++) {
+  for(int i = 0; i < NbLayer; i++) {
     ZonLayer[i] = (int)(3 * K1) + (int)((i + 1) * K2) + s->Num;
   }
   SurfLayer[0] = s->Num;
-  for(i = 0; i < NbLayer; i++) {
+  for(int i = 0; i < NbLayer; i++) {
     SurfLayer[i + 1] = (int)(2 * K1) + (int)((i + 1) * K2) + s->Num;
   }
 
@@ -509,46 +515,46 @@ static void Extrude_Surface3(void *data, void *dum)
   Tree_Action(s->Quadrangles, Extrude_Quadrangle_Phase3);
 }
 
-
-static void Extrude_Seg(Vertex * V1, Vertex * V2)
+static void Extrude_Seg(void *data, void *dum)
 {
-  int i, j, k;
-  Vertex *v1, *v2, *v3, *v4;
-  Simplex *s;
-  Quadrangle *q;
+  Simplex *ll = *(Simplex**)data;
+  Vertex *V1 = ll->V[0];
+  Vertex *V2 = ll->V[1];
 
   //printf("-curve vertex %d %p   %d %p\n", V1->Num, V1, V2->Num, V2);
 
-  k = 0;
-  for(i = 0; i <= NbLayer; i++) {
+  int k = 0;
+  for(int i = 0; i <= NbLayer; i++) {
     if(LineLayer[i]) {
+      Vertex *v1, *v2;
       List_Read(V1->Extruded_Points, k, &v1);
       List_Read(V2->Extruded_Points, k, &v2);
-      s = Create_Simplex(v1, v2, NULL, NULL);
+      Simplex *s = Create_Simplex(v1, v2, NULL, NULL);
       s->iEnt = LineLayer[i];
-      Tree_Add(THEV->Simp_Surf, &s);
+      Tree_Add(THEV->Lin_Surf, &s);
     }
-    for(j = 0; j < NbElmLayer[i]; j++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
       k++;
     }
   }
 
   k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
+  for(int i = 0; i < NbLayer; i++) {
+    for(int j = 0; j < NbElmLayer[i]; j++) {
+      Vertex *v1, *v2, *v3, *v4;
       List_Read(V1->Extruded_Points, k, &v1);
       List_Read(V2->Extruded_Points, k, &v2);
       List_Read(V1->Extruded_Points, k + 1, &v3);
       List_Read(V2->Extruded_Points, k + 1, &v4);
       if(SurfLayer[i]) {
         if(CTX.mesh.oldxtrude_recombine) {
-          q = Create_Quadrangle(v1, v2, v4, v3);
+          Quadrangle *q = Create_Quadrangle(v1, v2, v4, v3);
           q->iEnt = SurfLayer[i];
           Tree_Add(THEV->Quad_Surf, &q);
         }
         else {
           if(are_exist(v3, v2, Tree_Ares)) {
-            s = Create_Simplex(v3, v2, v1, NULL);
+            Simplex *s = Create_Simplex(v3, v2, v1, NULL);
             s->iEnt = SurfLayer[i];
             Tree_Add(THEV->Simp_Surf, &s);
             s = Create_Simplex(v3, v4, v2, NULL);
@@ -556,7 +562,7 @@ static void Extrude_Seg(Vertex * V1, Vertex * V2)
             Tree_Add(THEV->Simp_Surf, &s);
           }
           else {
-            s = Create_Simplex(v3, v4, v1, NULL);
+            Simplex *s = Create_Simplex(v3, v4, v1, NULL);
             s->iEnt = SurfLayer[i];
             Tree_Add(THEV->Simp_Surf, &s);
             s = Create_Simplex(v1, v4, v2, NULL);
@@ -573,116 +579,122 @@ static void Extrude_Seg(Vertex * V1, Vertex * V2)
 
 static void Extrude_Curve(void *data, void *dum)
 {
-  Curve **pC, *c;
-  Vertex *v1, *v2;
-  int i;
   if(!NbLayer)
     return;
-  pC = (Curve **) data;
-  c = *pC;
+
+  Curve *c = *(Curve **) data;
 
   if(c->Num < 0)
     return;
 
-  /* Numerotation automatique des entites physiques */
   Msg(INFO, "Extruding Curve %d", c->Num);
 
   LineLayer[0] = c->Num;
-  for(i = 0; i < NbLayer; i++) {
+  for(int i = 0; i < NbLayer; i++) {
     SurfLayer[i] = (int)(1 * K1) + (int)((i + 1) * K2) + c->Num;
     LineLayer[i + 1] = (int)(5 * K1) + (int)((i + 1) * K2) + c->Num;
   }
 
-  for(i = 0; i < List_Nbr(c->Vertices) - 1; i++) {
-    List_Read(c->Vertices, i, &v1);
-    List_Read(c->Vertices, i + 1, &v2);
-    Extrude_Seg(v1, v2);
-  }
-}
-
-static void Extrude_Pnt(Vertex * V1)
-{
-  int i, j, k;
-  Vertex *v1, *v2;
-  Simplex *s;
-
-  k = 0;
-  for(i = 0; i < NbLayer; i++) {
-    for(j = 0; j < NbElmLayer[i]; j++) {
-      List_Read(V1->Extruded_Points, k, &v1);
-      List_Read(V1->Extruded_Points, k + 1, &v2);
-      if(LineLayer[i]) {
-        s = Create_Simplex(v1, v2, NULL, NULL);
-        s->iEnt = LineLayer[i];
-        Tree_Add(THEV->Simp_Surf, &s);
-      }
-      k++;
-    }
-  }
-
+  List_Action(c->Vertices, Extrude_Vertex);
+  Tree_Action(c->Simplexes, Extrude_Seg);
 }
 
 static void Extrude_Point(void *data, void *dum)
 {
-  Vertex **pV, *v, **pV2;
-  int i;
-
   if(!NbLayer)
     return;
-  pV = (Vertex **) data;
-  v = *pV;
 
-  /* Numerotation automatique des entites physiques */
+  Vertex *v = *(Vertex **) data;
+
   Msg(INFO, "Extruding Vertex %d", v->Num);
-  for(i = 0; i < NbLayer; i++) {
+
+  for(int i = 0; i < NbLayer; i++) {
     LineLayer[i] = (int)(4 * K1) + (int)((i + 1) * K2) + v->Num;
   }
 
-  if((pV2 = (Vertex **) Tree_PQuery(THEM->Vertices, pV))) {
-    Extrude_Vertex(pV2, NULL);
-    Extrude_Pnt(*pV2);
+  // only extrude points for which a mesh vertex exists with the same
+  // number (i.e., start & end points of curves)
+  Vertex **pV;
+  if((pV = (Vertex **) Tree_PQuery(THEM->Vertices, &v))) {
+    Extrude_Vertex(pV, NULL);
+    int k = 0;
+    for(int i = 0; i < NbLayer; i++) {
+      for(int j = 0; j < NbElmLayer[i]; j++) {
+	Vertex *v1, *v2;
+	List_Read((*pV)->Extruded_Points, k, &v1);
+	List_Read((*pV)->Extruded_Points, k + 1, &v2);
+	if(LineLayer[i]) {
+	  Simplex *s = Create_Simplex(v1, v2, NULL, NULL);
+	  s->iEnt = LineLayer[i];
+	  Tree_Add(THEV->Lin_Surf, &s);
+	}
+	k++;
+      }
+    }
   }
-
+  
 }
 
 void FreeEP(void *a, void *b)
 {
-  Vertex *v = *((Vertex **) a);
-  Free_ExtrudedPoints(v->Extruded_Points);
-  v->Extruded_Points = NULL;
+  Vertex *v = *(Vertex **) a;
+  if(v->Extruded_Points){
+    Free_ExtrudedPoints(v->Extruded_Points);
+    v->Extruded_Points = NULL;
+  }
 }
 
 void Extrude_Mesh_Old(Mesh * M)
 {
-  int j;
-  Mesh MM;
-
   InitExtrudeParams();
-  LOCAL = &MM;
+
   THEM = M;
 
-  //clean up Extruded_Points stuff (in case another extrusion was
-  //performed before)
+  // clean up Extruded_Points stuff (in case another extrusion was
+  // performed before)
   Tree_Action(THEM->Vertices, FreeEP);
 
-  Create_BgMesh(WITHPOINTS, .2, LOCAL);
+  THEV = Create_Volume(1, MSH_VOLUME);
+  Tree_Add(M->Volumes, &THEV);
+
+  TRY_INVERSE = 0;
+
+ try_again:
+
+  if(Tree_Ares)
+    Tree_Delete(Tree_Ares);
+  if(Tree_Swaps)
+    Tree_Delete(Tree_Swaps);
+  Tree_Ares = Tree_Create(sizeof(nxn), compnxn);
+  Tree_Swaps = Tree_Create(sizeof(nxn), compnxn);
 
-  Tree_Left(M->Volumes, &THEV);
   Tree_Action(M->Surfaces, Extrude_Surface1);
 
   if(!CTX.mesh.oldxtrude_recombine) {
-    j = 0;
+    int j = 0;
     do {
-      TEST_IS_ALL_OK = 0;
+      BAD_TETS = 0;
       Tree_Action(M->Surfaces, Extrude_Surface2);
-      Msg(INFO, "%d swaps", TEST_IS_ALL_OK);
-      if(TEST_IS_ALL_OK == j)
-        break;
-      j = TEST_IS_ALL_OK;
-    } while(TEST_IS_ALL_OK);
+      Msg(INFO, "%d swaps", BAD_TETS);
+      if(BAD_TETS == j && j != 0) {
+	if(!TRY_INVERSE){
+	  Msg(WARNING, "Unable to swap all edges: trying again...");
+	  TRY_INVERSE = 1;
+	  goto try_again;
+	}
+	else{
+	  Msg(GERROR, "Unable to swap all edges (output mesh will be incorrect): use -recombine");
+	  break;
+	}
+      }
+      j = BAD_TETS;
+    } while(BAD_TETS);
   }
-
+  
   Tree_Action(M->Surfaces, Extrude_Surface3);
   Tree_Action(M->Curves, Extrude_Curve);
   Tree_Action(M->Points, Extrude_Point);
+
+  Tree_Delete(Tree_Ares);
+  Tree_Delete(Tree_Swaps);
 }
