@@ -1,7 +1,7 @@
 #ifndef _CUT_SPHERE_H_
-#define _CUT_SPHERE_H
+#define _CUT_SPHERE_H_
 
-// Copyright (C) 1997-2004 C. Geuzaine, J.-F. Remacle
+// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -30,6 +30,8 @@ extern "C"
 class GMSH_CutSpherePlugin : public GMSH_LevelsetPlugin
 {
   double levelset(double x, double y, double z, double val) const;
+  static double callback(int num, int action, double value, double *opt,
+			 double step, double min, double max);
 public:
   GMSH_CutSpherePlugin();
   void getName(char *name) const;
@@ -38,6 +40,14 @@ public:
   int getNbOptions() const;
   StringXNumber* getOption(int iopt);  
   Post_View *execute(Post_View *);
+
+  static double callbackX(int, int, double);
+  static double callbackY(int, int, double);
+  static double callbackZ(int, int, double);
+  static double callbackR(int, int, double);
+  static double callbackVol(int, int, double);
+  static double callbackRecur(int, int, double);
+  static void draw();
 };
 
 #endif
